import asyncio
import base64
from io import BytesIO
from concurrent.futures.process import BrokenProcessPool
import dataclasses
from pathlib import Path
from typing import Optional
import sys

import librosa
from nicegui import app, events, run, ui
import numpy as np
import plotly.graph_objects as go

from synth_mapping_helper.gui_tabs.utils import *
from synth_mapping_helper.utils import pretty_list, pretty_fraction, beat_to_second, second_to_beat
from synth_mapping_helper import synth_format, movement, analysis, audio_format, __version__, rails

PLOT_COLORS = {
    "right": "red",
    "left": "blue",
    "single": "green",
    "both": "orange",
    "combined": "black",
    "notes+rails": "black",
    "walls": "purple",
}

WARNING_MAX = 100  # Tab stops working if there are too many

def _in_slot(func, slot):
    def _handler():
        with slot:
            return func()
    return _handler

def _file_utils_tab() -> None:
    @dataclasses.dataclass
    class FileInfo:
        storage = app.storage.user
        data: Optional[synth_format.SynthFile] = None
        output_filename: str = "No file selected"
        output_bpm: float = 0.0
        output_offset: int = 0
        output_finalize: bool = False
        preview_audio: Optional[tuple[str, bytes]] = None
        # [diff][type]
        wall_densities: Optional[dict[str, dict[str, analysis.PlotDataContainer]]] = None
        # [diff][type][subtype]
        note_densities: Optional[dict[str, dict[str, dict[str, analysis.PlotDataContainer]]]] = None
        # [diff][type]
        running_counts: Optional[dict[str, dict[str, analysis.PlotDataContainer]]] = None
        # [diff][type]
        hand_curves: Optional[dict[str, dict[str, analysis.HAND_CURVE_TYPE]]] = None
        # [diff]
        warnings: Optional[dict[str, list[analysis.Warning]]] = None
        bpm_scan_data: Optional[dict] = None
        merged_filenames: list[str] = dataclasses.field(default_factory=list)

        @property
        def is_valid(self) -> bool:
            return self.data is not None

        def clear(self) -> None:
            self.data = None
            self.output_filename = "No base file selected"
            self.output_bpm = 0.0
            self.output_offset = 0
            self.merged_filenames = []
            self.bpm_scan_data = None
            self.wall_densities = None
            self.note_densities = None
            self.running_counts = None
            self.hand_curves = None
            self.warnings = None
            self.refresh()

        @handle_errors
        async def upload(self, e: events.UploadEventArguments) -> None:
            upl: ui.upload = e.sender  # type:ignore
            upl.reset()
            self.clear()
            if e.name.endswith(".synth"):
                self.data = try_load_synth_file(e)
                self.output_filename = add_suffix(e.name, "out")
            else:
                try:
                    self.data = synth_format.SynthFile.empty_from_audio(audio_file=BytesIO(e.content.read()), filename=e.name)
                except Exception as exc:
                    error(f"Creating .synth from '{e.name}' failed", exc=exc)
                    self.data = None
                else:
                    self.output_filename = self.data.meta.name + ".synth"
            if self.data is not None:
                self.output_bpm = self.data.bpm
                self.output_offset = self.data.offset_ms
                self.output_finalize = (self.data.bookmarks.get(0) == "#smh_finalized")
                self.merged_filenames = []
                self.bpm_scan_data = {"state": "Waiting"}
                ui.timer(0.1, self._calc_warn, once=True)
                ui.timer(0.2, self._calc_wden, once=True)
                ui.timer(0.3, self._calc_nden, once=True)
                ui.timer(0.4, self._calc_rcount, once=True)
                ui.timer(0.5, self._calc_hcurve, once=True)
                ui.timer(1.0, self._calc_bpm, once=True)

            self.refresh()

        @handle_errors
        def upload_merge(self, e: events.UploadEventArguments) -> None:
            upl: ui.upload = e.sender  # type:ignore
            upl.reset()
            if self.data is None:
                return
            merge = try_load_synth_file(e)
            if merge is None:
                return
            if merge.audio.raw_data != self.data.audio.raw_data:
                ui.notify("Difference in audio files detected. Merge may yield weird results.", type="warning")
            self.data.merge(merge, merge_bookmarks=merge_bookmarks.value)
            self.merged_filenames.append(e.name)
        
            self.refresh()

        def upload_cover(self, e: events.UploadEventArguments) -> None:
            upl: ui.upload = e.sender  # type:ignore
            upl.reset()
            if self.data is None:
                return
            if not e.name.lower().endswith(".png"):
                error("Cover image must be .png")
                return
            self.data.meta.cover_data = e.content.read()
            self.data.meta.cover_name = e.name
            ui.notify(f"Changed cover image to {e.name}", type="info")
            self.refresh()

        async def upload_audio(self, e: events.UploadEventArguments) -> None:
            upl: ui.upload = e.sender  # type:ignore
            upl.reset()
            if self.data is None:
                return
            try:
                raw_data = e.content.read()
                try:
                    # spinning up another process is slow, so lets attempt without conversion first
                    new_audio = audio_format.AudioData.from_raw(raw_data)
                except audio_format.AudioNotOggError as anoe:
                    ui.notify(f"Audio not ogg ({anoe.detected_format}), attempting conversion...", type="info")
                    # conversion takes a while, so offload that to another process
                    new_audio = await run.cpu_bound(audio_format.AudioData.from_raw, raw_data=raw_data, allow_conversion=True)
            except ValueError as ve:
                error("Error reading audio file", exc=ve, data=e.name)
            else:
                self.data.audio = new_audio
                self.data.meta.audio_name = e.name
                self._audio_info.refresh()
                ui.notify(f"Changed audio to {e.name}", type="info")
                self.bpm_scan_data = {"state": "Waiting"}
                # note: this callback runs inside the scope of the container of the uploader
                # so if we did a full refresh here, the info_card would get re-created, deleting the container and the timer with it
                # See https://github.com/zauberzeug/nicegui/issues/3187
                self._bpm_card.refresh()
                ui.timer(0.1, self._calc_bpm, once=True)

        @handle_errors
        def save(self) -> None:
            if self.data is None:
                return
            if self.output_bpm is None or not self.output_bpm > 0:
                error("BPM must be greater than 0", data=self.output_bpm)
                return
            if self.output_offset is None or self.output_offset < 0:
                error("Offset must be 0 or greater", data=self.output_offset)
                return
            if self.output_bpm != self.data.bpm:
                self.data.change_bpm(self.output_bpm)
            if self.output_offset != self.data.offset_ms:
                self.data.change_offset(self.output_offset)
            finalized = (self.data.bookmarks.get(0.0) == "#smh_finalized")
            if finalized != self.output_finalize:
                if not finalized:
                    self.data.bookmarks[0.0] = "#smh_finalized"
                    for _, diff_data in self.data.difficulties.items():
                        diff_data.apply_for_walls(movement.offset, offset_3d=(0,-2.1,0), types=synth_format.SLIDE_TYPES)
                else:
                    del self.data.bookmarks[0.0]
                    for _, diff_data in self.data.difficulties.items():
                        diff_data.apply_for_walls(movement.offset, offset_3d=(0,2.1,0), types=synth_format.SLIDE_TYPES)

            data = BytesIO()
            self.data.save_as(data)
            ui.download.content(data.getvalue(), filename=self.output_filename or "unnamed.synth")

        def save_errors(self):
            out = [
                "SMH-GUI fixed error report:",
                f"  SMH Version: {__version__}",
                f"  Base BPM: {self.data.bpm}",
                f"  Base Offset: {self.data.offset_ms}",
                f"  Merged {len(self.merged_filenames)} other files",
            ]
            for diff, errors in self.data.errors.items():
                for jpe, time in errors:
                    out.append(f"{diff}@{time}: {jpe!r}")

            ui.download.content('\n'.join(out).encode(), filename="smh_error_report.txt")
            info("Saved error log")

        def refresh(self) -> None:
            self.info_card.refresh()
            self.stats_card.refresh()

        @handle_errors
        async def add_silence(self, before_start_ms: int=0, after_end_ms: int=0) -> None:
            if self.data is None:
                return
            self.bpm_scan_data = {"state": "Waiting"}
            self._bpm_card.refresh()
            self.data = await run.cpu_bound(self.data.with_added_silence, before_start_ms=before_start_ms, after_end_ms=after_end_ms)
            self._audio_info.refresh()
            ui.timer(0.01, self._calc_bpm, once=True)

        @ui.refreshable
        def _audio_info(self) -> None:
            if self.data is None:
                return
            default_source = "data:audio/ogg;base64,"+base64.b64encode(self.data.audio.raw_data).decode()
            preview_audio = ui.audio(default_source)
            with ui.row():
                with ui.number("BPM", min=1.0, max=600.0, step=0.1).props("dense").classes("w-20").bind_value(self, "output_bpm"):
                    ui.tooltip("").bind_text_from(self, "output_bpm", backward=lambda bpm: f"{round(60000/bpm)} ms/b" if bpm is not None else "Invalid")
                def _multiply_bpm(mult: float) -> None:
                    self.output_bpm = round(self.output_bpm*mult, 3)
                ui.button("½", on_click=lambda _: _multiply_bpm(1/2), color="secondary").props("dense outline").classes("w-8 my-auto").tooltip("Divide BPM by 2")
                ui.button("2", on_click=lambda _: _multiply_bpm(2)).props("dense outline").classes("w-8 my-auto").tooltip("Double BPM")
                ui.separator().props("vertical")
                ui.button("⅓", on_click=lambda _: _multiply_bpm(1/3), color="secondary").props("dense outline").classes("w-8 my-auto").tooltip("Divide BPM by 3")
                ui.button("3", on_click=lambda _: _multiply_bpm(3)).props("dense outline").classes("w-8 my-auto").tooltip("Triple BPM")
            with ui.row():
                with ui.number("Offset", min=0, step=1, suffix="ms").props("dense").classes("w-20").bind_value(self, "output_offset"):
                    def _update_offset_tooltip(_) -> str:
                        bpm = self.output_bpm
                        offset = self.output_offset
                        return f"-{round(60000/bpm-offset)} ms" if bpm and offset is not None else "Invalid"
                    ui.tooltip("").bind_text_from(self, "output_bpm", backward=_update_offset_tooltip).bind_text_from(self, "output_offset", backward=_update_offset_tooltip)
                def _shift_offset(beats: float) -> None:
                    beat_time = 60/self.output_bpm
                    offset = round((self.output_offset/1000 + beats*beat_time)*1000)
                    if offset >= 0:
                        self.output_offset = offset
                    else:
                        ui.notify("Negative offset is not supported", type="warning")
                ui.button(icon="chevron_left", on_click=lambda _: _shift_offset(-1), color="negative").props("dense outline").classes("w-8 my-auto").tooltip("Subtract a beat from offset")
                ui.button(icon="chevron_right", on_click=lambda _: _shift_offset(1), color="positive").props("dense outline").classes("w-8 my-auto").tooltip("Add a beat to offset")
                ui.separator().props("vertical")
                ui.button("<½", on_click=lambda _: _shift_offset(-0.5), color="negative").props("dense outline").classes("w-8 my-auto").tooltip("Subtract half a beat from offset")
                ui.button(">½", on_click=lambda _: _shift_offset(0.5), color="positive").props("dense outline").classes("w-8 my-auto").tooltip("Add half a beat to offset")
            with ui.row():
                def _reset_bpm():
                    self.output_bpm = self.data.bpm
                    self.output_offset = self.data.offset_ms
                reset_button = ui.button(icon="undo", on_click=_reset_bpm, color="warning").props("dense outline").classes("w-8 my-auto").tooltip("Reset BPM and Offset to original values")
                ui.separator().props("vertical")

                async def _pad_offset() -> None:
                    offset_ms = int(self.output_offset)
                    ui.notify(f"Adding {offset_ms} ms of silence. This may take a few seconds.", type="info")
                    self.output_offset = 0
                    await self.add_silence(before_start_ms=offset_ms)
                pad_button = ui.button(icon="keyboard_tab", on_click=_pad_offset, color="positive").props("dense").classes("w-8 my-auto").bind_enabled_from(self, "output_offset")
                with pad_button:
                    ui.tooltip().bind_text_from(self, "output_offset", lambda o: f"Add {int(o) if o else '<offset>'} ms of silence before the audio")
                with ui.dialog() as pad_dialog, ui.card():
                    ui.label("Note: After these, offset should be re-applied")
                    with ui.row():
                        before_ms = ui.number("Before", value=0, suffix="ms").props("dense").classes("w-24").tooltip("Amount to add. Negative trims instead.")
                        async def _pad_audio():
                            pad_dialog.close()
                            ui.notify(f"Padding audio. This may take a few seconds.", type="info")
                            await self.add_silence(before_start_ms=before_ms.value, after_end_ms=after_ms.value)
                            await self._calc_bpm
                        ui.button(icon="settings_ethernet", color="positive", on_click=_pad_audio).props("dense").classes("my-auto").tooltip("Add silence to start and end")
                        after_ms = ui.number("After", value=0, suffix="ms").props("dense").classes("w-24").tooltip("Amount to add. Negative trims instead.")
                    ui.separator()
                    with ui.row():
                        async def _trim_silence():
                            pad_dialog.close()
                            ui.notify(f"Trimming silence from audio. This may take a few seconds.", type="info")
                            trim_start, trim_end = await run.io_bound(audio_format.find_trims, raw_audio_data=self.data.audio.raw_data)
                            await self.add_silence(before_start_ms=-round(trim_start*1000), after_end_ms=-round(trim_end*1000))
                            ui.notify(f"Trimmed {trim_start:.3f} s from beginning and {trim_end:.3f} s from end", type="info")
                        ui.button(icon="content_cut volume_off", color="negative", on_click=_trim_silence).props("dense").classes("w-16").tooltip("Detect and remove silence from beginning and end")
                        async def _add_2s():
                            pad_dialog.close()
                            ui.notify(f"Adding two seconds of silence. This may take a few seconds.", type="info")
                            await self.add_silence(before_start_ms=2000)
                        ui.button("2s", icon="start", color="positive", on_click=_add_2s).props("dense").tooltip("Add two seconds of silence before start")
                        async def _align_bookmark():
                            pad_dialog.close()
                            bookmarks = sorted(self.data.bookmarks.items())
                            if not bookmarks:
                                error(msg="No bookmarks found")
                            bm_beat, bm_name = bookmarks[0]
                            bm_time = beat_to_second(bm_beat, bpm=self.data.bpm) - self.data.offset_ms/1000
                            if bm_time > 2:
                                ui.notify(f"Removing {bm_time-2:.3f} s to align first bookmark ({bm_name!r}) to 2 second mark. This may take a few seconds.", type="info")
                            elif bm_time < 2:
                                ui.notify(f"Adding {2-bm_time:.3f} s to align first bookmark ({bm_name!r}) to 2 second mark. This may take a few seconds.", type="info")
                            before_start_ms = round((2-bm_time)*1000)
                            await self.add_silence(before_start_ms=before_start_ms)
                            self.data.change_offset(0)
                            self.output_offset = 0
                        with ui.button("2s", icon="bookmark_added", color="positive", on_click=_align_bookmark).props("dense").bind_enabled_from(self.data, "bookmarks"):
                            ui.tooltip().bind_text_from(self.data, "bookmarks", backward=lambda b: "Align first bookmark with 2 second mark" + ("" if b else " (no bookmarks found)"))
                def _open_pad_dialog() -> None:
                    pad_dialog.open()
                    pad_dialog.update()
                ui.button(icon="settings_ethernet", color="info", on_click=_open_pad_dialog).props("dense").classes("w-8 my-auto").tooltip("Pad or trim audio")
                ui.separator().props("vertical")
                @handle_errors
                async def _add_clicks(e: events.ClickEventArguments):
                    if self.data is None:
                        return
                    bpm = self.output_bpm
                    offset = self.output_offset
                    if bpm is None or not bpm > 0:
                        error("BPM must be greater than 0", data=bpm)
                        return
                    if offset is None or offset < 0:
                        error("Offset must be 0 or greater", data=offset)
                        return
                    btn: ui.button = e.sender  # type: ignore
                    btn.props('color="grey"').classes("cursor-wait")  # turn grey and indicate wait
                    data = await run.cpu_bound(audio_format.audio_with_clicks, raw_audio_data=self.data.audio.raw_data, duration=self.data.audio.duration, bpm=bpm, offset_ms=offset)
                    preview_audio.set_source("data:audio/ogg;base64,"+base64.b64encode(data).decode())
                    btn.props('color="positive"').classes(remove="cursor-wait")  # reset visuals
                ui.button(icon="timer", on_click=_add_clicks, color="positive").props("dense outline").classes("w-8 my-auto").tooltip("Add or update clicks in preview")
                ui.button(icon="timer_off", on_click=lambda _: preview_audio.set_source(default_source), color="negative").props("dense outline").classes("w-8 my-auto").tooltip("Remove clicks from preview")
        @ui.refreshable
        def info_card(self):
            if self.data is None:
                ui.label("Load a map to show info")
                return
            ui.markdown("**Edit metadata**")
            ui.separator()
            meta = self.data.meta
            with ui.row():
                with ui.column().classes("w-40"):
                    ui.input("Name").props("dense").classes("h-8").bind_value(meta, "name")
                    ui.input("Artist").props("dense").classes("h-8").bind_value(meta, "artist")
                    ui.input("Mapper").props("dense").classes("h-8").bind_value(meta, "mapper")
                    ui.checkbox("Explicit lyrics").classes("h-8").props("dense").bind_value(meta, "explicit")
                ui.separator().props("vertical")
                with ui.upload(label="Replace Cover" if meta.cover_data else "Set Cover", auto_upload=True, on_upload=self.upload_cover).classes("w-32").props('accept="image/png"').add_slot("list"):
                    ui.image("data:image/png;base64,"+base64.b64encode(meta.cover_data).decode()).tooltip(meta.cover_name)
                ui.separator().props("vertical")
                with ui.upload(label="Edit / Replace Audio", auto_upload=True, on_upload=self.upload_audio).props('accept="audio/ogg,*/*"').classes("w-80").add_slot("list"):
                    self._audio_info()

        @handle_errors
        async def _calc_bpm(self):
            sd = self.bpm_scan_data  # create a pointer to avoid messing up data when it gets replaced during calc
            if app.storage.user.get("fileutils_bpm_error_count", 0) >= 3:
                ui.notify("BPM calculation skipped due to persistent errors. I am not sure what causes this, and it only affects some computers...", type="warning")
                sd["state"] = "Disabled"
                self._bpm_card.refresh()
                return
            try:
                sd["state"] = "Processing audio..."
                self._bpm_card.refresh()
                sd["bpm_override"] = None
                sd |= await run.cpu_bound(analysis.bpm_aio, raw_data=self.data.audio.raw_data)
                sd["state"] = "Done"
                self._bpm_card.refresh()
            except:
                sd["state"] = "Error"
                app.storage.user["fileutils_bpm_error_count"] = app.storage.user.get("fileutils_bpm_error_count", 0) + 1
                self._bpm_card.refresh()
                raise
            # success: clear error counter
            app.storage.user["fileutils_bpm_error_count"] = 0

        async def _recalc_beats(self):
            # have this seperate so it could be updated seperately later
            sd = self.bpm_scan_data
            sr = sd["sr"]
            onsets = sd["onsets"]
            bpm_override = sd["bpm_override"]
            if bpm_override is None:
                bpm_sections = sd["bpm_sections"]
            else:
                bpm_sections = [(0, onsets.shape[-1], bpm_override, 1)]
            sd["state"] = "Recalculating offset"
            self._bpm_card.refresh()
            sd["offset_sections"] = await run.cpu_bound(analysis.find_offsets, onsets=onsets, sr=sr, bpm_sections=bpm_sections)
            sd["state"] = "Done"
            self._bpm_card.refresh()

        @ui.refreshable
        def _bpm_card(self) -> None:
            if self.bpm_scan_data is None:
                ui.label("No BPM data")
                return
            if self.bpm_scan_data["state"] != "Done":
                if self.bpm_scan_data["state"] in ("Error", "Disabled"):
                    ui.label(f'BPM Calcuation {self.bpm_scan_data["state"]} (Last {app.storage.user["fileutils_bpm_error_count"]} attempts failed). You may retry below').classes("my-auto")
                    @handle_errors
                    async def _retry_bpm_calc():
                        if app.storage.user["fileutils_bpm_error_count"] >=3:
                            app.storage.user["fileutils_bpm_error_count"] = 0
                        self.bpm_scan_data = {"state": "Waiting"}
                        self._bpm_card.refresh()
                        await self._calc_bpm()
                    ui.button("Retry", on_click=_retry_bpm_calc)
                    return
                with ui.row():
                    ui.spinner(size="xl")
                    ui.label().classes("my-auto").bind_text_from(self.bpm_scan_data, "state", backward=lambda s: s + " (this may take a few seconds)")
                return
            ui.label("BPM Scan Results")
            sr = self.bpm_scan_data["sr"]
            onsets = self.bpm_scan_data["onsets"]
            peak_values = self.bpm_scan_data["peak_values"]
            pulse = self.bpm_scan_data["pulse"]
            peak_bpms = self.bpm_scan_data["peak_bpms"]
            best_bpm = self.bpm_scan_data["best_bpm"]
            bpm_sections = self.bpm_scan_data["bpm_sections"]
            offset_sections = self.bpm_scan_data["offset_sections"]
            bpm_override = self.bpm_scan_data["bpm_override"]

            with ui.row():
                async def _reset_bpm():
                    self.bpm_scan_data["bpm_override"] = None
                    await self._recalc_beats()
                if bpm_override is not None:
                    ui.button(icon="undo", on_click=_reset_bpm, color="warning").props("dense outline").classes("my-auto").tooltip("Reset back to detected BPM")
                async def _override_bpm():
                    self.bpm_scan_data["bpm_override"] = self.output_bpm
                    await self._recalc_beats()
                ui.button("BPM Override", icon="south", on_click=_override_bpm, color="warning").props("dense outline").tooltip("Override detected BPM with current BPM and recalculate beats and offset")
                with ui.dropdown_button("Apply", auto_close=True, icon="auto_fix_high").props("dense outline").tooltip("Apply detected BPM and offset"):
                    for i, (_, _, _, section_bpm, section_offset, _) in enumerate(offset_sections):
                        color = "green" if section_bpm==best_bpm else ("blue", "red")[i%2]
                        def _apply_bpm(bpm=section_bpm, offset_ms=section_offset):
                            self.output_bpm = bpm
                            self.output_offset = offset_ms
                        ui.item(f"Section {i+1}: {section_bpm} BPM, {section_offset} ms", on_click=_apply_bpm).classes(f"text-{color}")
                def _add_bookmarks():
                    # clear existing bpm bookmarks
                    for t, n in list(self.data.bookmarks.items()):
                        if n.startswith("#smh_bpm"):
                            del self.data.bookmarks[t]
                    # add new
                    self.data.bookmarks |= {
                        librosa.frames_to_time(s, sr=sr): f"#smh_bpm: {s_bpm}"
                        for (s, _, s_bpm, _) in bpm_sections
                    }
                    # update plots
                    self.stats_card.refresh()
                ui.button("Add bookmarks", icon="bookmark", on_click=_add_bookmarks, color="positive").props("dense outline").tooltip("Add bookmark on section starts")
            ui.markdown("""
                Note: This may not be accurate. Double check below if the *Error* plot below for the section stays low (meaning *Actual Beats* and *Stable BPM* align), and this matches up with the *Note onsets* plot.

                You can zoom the plots by dragging the axis or selecting a rectangle. Use the "Reset Axis" button (home icon) in the top right of each plot to reset. You can also switch to "Pan" mode there. 
            """)

            bpmfig = go.Figure(
                layout=go.Layout(
                    xaxis=go.layout.XAxis(title="Time"),
                    yaxis=go.layout.YAxis(title="BPM"),
                    legend=go.layout.Legend(x=0, xanchor="left", y=1, yanchor="top", orientation="h", groupclick="toggleitem"),
                    margin=go.layout.Margin(l=0, r=0, t=0, b=0),
                    hovermode="x unified",
                ),
            )
            bpmfig.add_scattergl(
                x=librosa.times_like(peak_bpms, sr=sr), y=peak_bpms,
                name="BPM",
            )
            for i, (section_start, section_end, section_bpm, str_sum) in enumerate(bpm_sections):
                color = "green" if section_bpm==best_bpm else ("blue", "red")[i%2]
                start_time, end_time = librosa.frames_to_time(section_start, sr=sr), librosa.frames_to_time(section_end, sr=sr)
                bpmfig.add_shape(
                    dict(type="rect", x0=librosa.frames_to_time(section_start, sr=sr), x1=librosa.frames_to_time(section_end, sr=sr), y0=0, y1=str_sum),
                    line_width=0, fillcolor=color, opacity=0.1, yref="paper",
                )
                bpmfig.add_annotation(
                    go.layout.Annotation(text=f"{section_bpm}", yanchor="bottom", font=dict(color=color)),
                    yref="paper", showarrow=False,
                    x=(start_time+end_time)/2, y=0,
                )
            ui.plotly(bpmfig).classes("w-full h-96")

            onset_fig = go.Figure(
                layout=go.Layout(
                    xaxis=go.layout.XAxis(title="Time"),
                    yaxis=go.layout.YAxis(title=" "),
                    legend=go.layout.Legend(x=0, xanchor="left", y=1, yanchor="top", orientation="h", groupclick="toggleitem", bgcolor="rgba(255,255,255,0.3)", borderwidth=1),
                    margin=go.layout.Margin(l=0, r=0, t=0, b=0),
                    hovermode="x unified",
                ),
            )
            onset_fig.add_scattergl(
                x=librosa.times_like(onsets, sr=sr), y=onsets,
                name="Note onsets",
                legendgroup="common",
                legendgrouptitle=dict(text="Common"),
            )
            onset_fig.add_scattergl(
                x=librosa.times_like(peak_values, sr=sr), y=peak_values,
                name="BPM confidence",
                legendgroup="common",
            )
            onset_fig.add_scattergl(
                x=librosa.times_like(pulse, sr=sr), y=pulse,
                name="Pulse curve",
                visible="legendonly",  # hide by default
                legendgroup="common",
            )
            for i, (section_start, section_end, beats, section_bpm, section_offset, offset_error) in enumerate(offset_sections):
                color = "green" if section_bpm==best_bpm else ("blue", "red")[i%2]
                start_time, end_time = librosa.frames_to_time(section_start, sr=sr), librosa.frames_to_time(section_end, sr=sr)
                onset_fig.add_vrect(
                    start_time, end_time,
                    line_width=0, fillcolor=color, opacity=0.1,
                    annotation=go.layout.Annotation(text=f"Section {i+1}<br>{section_bpm} bpm<br>{section_offset} ms", yanchor="bottom", yref="paper", font=dict(color=color), bgcolor="white"),
                    annotation_position="bottom",
                )
                onset_fig.add_scattergl(
                    # just vertical lines
                    x=beats.repeat(3), y=[0,1,None]*len(beats),
                    name="Actual Beats",
                    line=dict(dash="dash", color=color),
                    mode="lines",
                    visible="legendonly",  # hide by default
                    legendgroup=f"sec_{i+1}",
                    legendgrouptitle=dict(text=f"Section {i+1}", font=dict(color=color))
                )
                beat_time = 60/section_bpm
                stable_beats = np.arange((start_time-start_time%beat_time)-(section_offset/1000)%beat_time+beat_time, end_time, beat_time)
                onset_fig.add_scattergl(
                    # just vertical lines
                    x=stable_beats.repeat(3), y=[0,1,None]*len(stable_beats),
                    name="Stable BPM",
                    line=dict(dash="dot", color=color),
                    mode="lines",
                    visible="legendonly",  # hide by default
                    legendgroup=f"sec_{i+1}",
                )
                onset_fig.add_scattergl(
                    x=beats,
                    y=offset_error,
                    name="Offset Error",
                    line=dict(color=color),
                    mode="lines",
                    legendgroup=f"sec_{i+1}",
                )
            ui.plotly(onset_fig).classes("w-full h-96")

        @handle_errors
        async def _calc_wden(self):
            self.wall_densities = await run.cpu_bound(analysis.all_wall_densities, diffs=self.data.difficulties)
            self._density_card.refresh()

        @handle_errors
        async def _calc_nden(self):
            self.note_densities = await run.cpu_bound(analysis.all_note_densities, diffs=self.data.difficulties)
            self._density_card.refresh()

        @handle_errors
        async def _calc_rcount(self):
            self.running_counts = await run.cpu_bound(analysis.all_running_counts, diffs=self.data.difficulties)
            self._density_card.refresh()

        @handle_errors
        async def _calc_hcurve(self):
            self.hand_curves = await run.cpu_bound(analysis.all_hand_curves, diffs=self.data.difficulties)
            self._hands_card.refresh()

        @handle_errors
        async def _calc_warn(self):
            self.warnings = await run.cpu_bound(
                analysis.all_warnings,
                diffs=self.data.difficulties,
                last_beat=second_to_beat(self.data.audio.duration-self.data.offset_ms/1000, bpm=self.data.bpm)
            )
            self._stats_table.refresh()
            self._warnings_card.refresh()

        def _wden_content(self, den_dict: dict[str, analysis.PlotDataContainer]) -> None:
            wfig = go.Figure(
                layout=go.Layout(
                    yaxis=go.layout.YAxis(title="Visible Walls (4s)"),
                    legend=go.layout.Legend(x=0, xanchor="left", y=1, yanchor="bottom", orientation="h"),
                    margin=go.layout.Margin(l=0, r=0, t=0, b=0),
                    hovermode="x unified",
                ),
            )
            for t, b in self.data.bookmarks.items():
                wfig.add_vline(t, line={"color": "lightgray", "dash": "dash"}, annotation=go.layout.Annotation(text="🔖", font=dict(color="gray"), hovertext=b, xanchor="center", yanchor="bottom"), annotation_position="bottom")

            # show horizontal lines when combined y is close to or over the limit
            max_com_d = den_dict["combined"].max_value
            if max_com_d > 0.9 * analysis.QUEST_WIREFRAME_LIMIT:
                wfig.add_hline(analysis.QUEST_WIREFRAME_LIMIT, line={"color": "gray", "dash": "dash"}, annotation=go.layout.Annotation(text="Quest wireframe (combined)", xanchor="left", yanchor="bottom"), annotation_position="left")
            if max_com_d > 0.9 * analysis.QUEST_RENDER_LIMIT:
                wfig.add_hline(analysis.QUEST_RENDER_LIMIT, line={"color": "red", "dash": "dash"}, annotation=go.layout.Annotation(text="Quest limit (combined)", xanchor="left", yanchor="bottom"), annotation_position="left")
            # show horizontal lines when single y is over the limit
            max_single_d = max(pdc.max_value for wt, pdc in den_dict.items() if wt != "combined")
            if max_single_d > 0.95 * analysis.PC_TYPE_DESPAWN:
                wfig.add_hline(analysis.PC_TYPE_DESPAWN, line={"color": "yellow", "dash": "dash"}, annotation=go.layout.Annotation(text="PC despawn (per type)", xanchor="left", yanchor="bottom"), annotation_position="left")

            for wt in ("combined", *synth_format.WALL_TYPES):
                pdc = den_dict[wt]
                if pdc.max_value:
                    wfig.add_scattergl(
                        x=pdc.plot_data[:,0], y=pdc.plot_data[:,1], name=f"{wt} [{analysis.wall_mode(pdc.max_value, combined=(wt == 'combined'))}]",
                        showlegend=True,
                        # start with only combined visible and single only when above PC limit
                        visible=(wt == "combined" or pdc.max_value > 0.95 * analysis.PC_TYPE_DESPAWN) or "legendonly"
                    )
            ui.plotly(wfig).classes("w-full h-96")

        def _nden_content(self, den_dict: dict[str, dict[str, analysis.PlotDataContainer]]) -> None:
            # mostly the same thing as walls, but for combined notes and rail nodes
            nfig = go.Figure(
                layout=go.Layout(
                    yaxis=go.layout.YAxis(title="Visible (4s)"),
                    legend=go.layout.Legend(x=-0.05, xanchor="right", y=1, yanchor="top", orientation="v", groupclick="toggleitem"),
                    margin=go.layout.Margin(l=0, r=0, t=0, b=0),
                    hovermode="x unified",
                ),
            )
            for t, b in self.data.bookmarks.items():
                nfig.add_vline(t, line={"color": "lightgray", "dash": "dash"}, annotation=go.layout.Annotation(text="🔖", font=dict(color="gray"), hovertext=b, xanchor="center", yanchor="bottom"), annotation_position="bottom")

            for nt in ("combined", *synth_format.NOTE_TYPES):
                den_subdict = den_dict[nt]
                for sub_t, pdc in den_subdict.items():
                    if pdc.max_value:
                        nfig.add_scattergl(
                            x=pdc.plot_data[:,0], y=pdc.plot_data[:,1], name=f"{nt} {sub_t}s [max {round(pdc.max_value)}]",
                            showlegend=True,
                            legendgroup=nt,
                            line={"color": PLOT_COLORS[nt]},
                            # start with only combined note visible
                            visible=(nt == "combined" and sub_t == "note") or "legendonly",
                        )
            ui.plotly(nfig).classes("w-full h-128")

        def _rcount_content(self, rcount_dict: dict[str, analysis.PlotDataContainer]) -> None:
            # mostly the same thing as densities, but for total notes and walls
            nfig = go.Figure(
                layout=go.Layout(
                    yaxis=go.layout.YAxis(title="Count"),
                    legend=go.layout.Legend(x=-0.05, xanchor="right", y=1, yanchor="top", orientation="v", groupclick="toggleitem"),
                    margin=go.layout.Margin(l=0, r=0, t=0, b=0),
                    hovermode="x unified",
                ),
            )
            for t, b in self.data.bookmarks.items():
                nfig.add_vline(t, line={"color": "lightgray", "dash": "dash"}, annotation=go.layout.Annotation(text="🔖", font=dict(color="gray"), hovertext=b, xanchor="center", yanchor="bottom"), annotation_position="bottom")

            for nt in ("notes+rails", "walls", *synth_format.ALL_TYPES):
                pdc = rcount_dict[nt]
                if pdc.max_value:
                    nfig.add_scattergl(
                        x=pdc.plot_data[:,0], y=pdc.plot_data[:,1], name=nt,
                        showlegend=True,
                        legendgroup=nt,
                        line={"color": PLOT_COLORS.get(nt, "grey")},
                        # start with only combined note visible
                        visible=(nt in ("notes+rails", "walls")) or "legendonly",
                    )
            ui.plotly(nfig).classes("w-full h-128")


        def _hcurve_content(self, curves: dict[str, analysis.HAND_CURVE_TYPE]|None, warnings: list[analysis.Warning]|None, diff_data: synth_format.DataContainer) -> None:
            xfig = go.Figure(
                layout=go.Layout(
                    yaxis=go.layout.YAxis(title="X: right (+) <-> left (-)", range=(7,-7)),
                    legend=go.layout.Legend(x=0, xanchor="left", y=1, yanchor="bottom", orientation="h"),
                    margin=go.layout.Margin(l=0, r=0, t=0, b=0),
                    hovermode="x unified",
                ),
            )
            yfig = go.Figure(
                layout=go.Layout(
                    yaxis=go.layout.YAxis(title="Y: down (-) <-> up (+)", range=(-5,5)),
                    legend=go.layout.Legend(x=0, xanchor="left", y=1, yanchor="bottom", orientation="h"),
                    margin=go.layout.Margin(l=0, r=0, t=0, b=0),
                    hovermode="x unified",
                ),
            )
            vfig = go.Figure(
                layout=go.Layout(
                    yaxis=go.layout.YAxis(title="Velocity (m/s)"),
                    legend=go.layout.Legend(x=0, xanchor="left", y=1, yanchor="bottom", orientation="h"),
                    margin=go.layout.Margin(l=0, r=0, t=0, b=0),
                    hovermode="x unified",
                ),
            )
            afig = go.Figure(
                layout=go.Layout(
                    yaxis=go.layout.YAxis(title="Acceleration (m/s²)"),
                    legend=go.layout.Legend(x=0, xanchor="left", y=1, yanchor="bottom", orientation="h"),
                    margin=go.layout.Margin(l=0, r=0, t=0, b=0),
                    hovermode="x unified",
                ),
            )
            for t, b in self.data.bookmarks.items():
                for f in (xfig, yfig, vfig, afig):
                    f.add_vline(t, line={"color": "lightgray", "dash": "dash"}, annotation=go.layout.Annotation(text="🔖", font=dict(color="gray"), hovertext=b, xanchor="center", yanchor="bottom"), annotation_position="bottom")

            if curves is not None:
                any_vel = False
                vel_mult = synth_format.GRID_SCALE / beat_to_second(1/analysis.CURVE_INTERP, self.data.bpm)
                any_acc = False
                acc_mult = synth_format.GRID_SCALE / beat_to_second(1/analysis.CURVE_INTERP, self.data.bpm)
                for nt, (pos, vel, acc) in curves.items():
                    xfig.add_scattergl(
                        x=pos[:,2], y=pos[:,0], name="curve",
                        showlegend=True,
                        legendrank=1,
                        line={"color": PLOT_COLORS[nt], "width": 1}
                    )
                    yfig.add_scattergl(
                        x=pos[:,2], y=pos[:,1], name="curve",
                        showlegend=True,
                        legendrank=1,
                        line={"color": PLOT_COLORS[nt], "width": 1}
                    )
                    if not np.isnan(pos).all():
                        note_list = []
                        rail_list = []
                        node_list = []
                        for _, nodes in diff_data.get_object_dict(nt).items():
                            note_list.append(nodes[:1])
                            if nodes.shape[0] > 1:
                                node_list.append(nodes[1:])
                                rail_list.append(rails.interpolate_nodes(nodes, "spline", 1/analysis.CURVE_INTERP))
                                rail_list.append(np.full((1,3,), np.nan))
                        note_array = np.concatenate(note_list)
                        xfig.add_scattergl(
                            x=note_array[:,2], y=note_array[:,0], name="notes",
                            showlegend=True,
                            legendrank=2,
                            mode="markers",
                            marker={"color": PLOT_COLORS[nt], "size": 8},
                        )
                        yfig.add_scattergl(
                            x=note_array[:,2], y=note_array[:,1], name="notes",
                            showlegend=True,
                            legendrank=2,
                            mode="markers",
                            marker={"color": PLOT_COLORS[nt], "size": 8},
                        )
                        if rail_list:
                            rail_array = np.concatenate(rail_list)
                            xfig.add_scattergl(
                                x=rail_array[:,2], y=rail_array[:,0], name="rails",
                                showlegend=True,
                                legendrank=2,
                                line={"color": PLOT_COLORS[nt], "width": 2}
                            )
                            yfig.add_scattergl(
                                x=rail_array[:,2], y=rail_array[:,1], name="rails",
                                showlegend=True,
                                legendrank=2,
                                line={"color": PLOT_COLORS[nt], "width": 2}
                            )
                            node_array = np.concatenate(node_list)
                            xfig.add_scattergl(
                                x=node_array[:,2], y=node_array[:,0], name="nodes",
                                showlegend=True,
                                legendrank=2,
                                mode="markers",
                                marker={"color": PLOT_COLORS[nt], "size": 4},
                            )
                            yfig.add_scattergl(
                                x=node_array[:,2], y=node_array[:,1], name="nodes",
                                showlegend=True,
                                legendrank=2,
                                mode="markers",
                                marker={"color": PLOT_COLORS[nt], "size": 4},
                            )
                    if not np.isnan(vel).all():
                        v_mag = np.sqrt((vel[:,0]**2) + (vel[:,1]**2)) * vel_mult
                        vfig.add_scattergl(
                            x=vel[:,2], y=v_mag, name="total",
                            showlegend=True,
                            legendrank=1,
                            line={"color": PLOT_COLORS[nt]}
                        )
                        vfig.add_scattergl(
                            x=vel[:,2], y=vel[:,0] * vel_mult, name="x",
                            showlegend=True,
                            legendrank=2,
                            line={"color": PLOT_COLORS[nt]},
                            visible="legendonly",
                        )
                        vfig.add_scattergl(
                            x=vel[:,2], y=vel[:,1] * vel_mult, name="y",
                            showlegend=True,
                            legendrank=2,
                            line={"color": PLOT_COLORS[nt]},
                            visible="legendonly",
                        )
                        any_vel = True
                    if not np.isnan(acc).all():
                        a_mag = np.sqrt((acc[:,0]**2) + (acc[:,1]**2)) * acc_mult
                        afig.add_scattergl(
                            x=acc[:,2], y=a_mag, name="total",
                            showlegend=True,
                            legendrank=1,
                            line={"color": PLOT_COLORS[nt]}
                        )
                        afig.add_scattergl(
                            x=acc[:,2], y=acc[:,0] * acc_mult, name="x",
                            showlegend=True,
                            legendrank=2,
                            line={"color": PLOT_COLORS[nt]},
                            visible="legendonly",
                        )
                        afig.add_scattergl(
                            x=acc[:,2], y=acc[:,1] * acc_mult, name="y",
                            showlegend=True,
                            legendrank=2,
                            line={"color": PLOT_COLORS[nt]},
                            visible="legendonly",
                        )
                        any_acc = True
            if warnings is not None:
                warning_types = self.storage.get("fileutils_warnings_types")
                warnings = [w for w in warnings if w.type in warning_types]
                if len(warnings) > WARNING_MAX:
                    ui.label(f"Too many warnings ({len(warnings)}), marking first {WARNING_MAX} only. See the warnings table for the rest.")
                for w in warnings[:WARNING_MAX]:
                    color = PLOT_COLORS.get(w.note_type, "black")
                    for fn, fig in zip("xy", (xfig, yfig)):
                        if fn in w.figure:
                            if w.start_beat == w.end_beat:
                                fig.add_vline(
                                    w.start_beat,
                                    line_color=color, opacity=0.2,
                                    annotation=go.layout.Annotation(
                                        text=w.icon, yanchor="top", yref="paper",
                                        hovertext=f"{w.note_type} {w.note_rail} @ {pretty_fraction(w.start_beat)}<br>{w.text}",
                                    ),
                                    annotation_position="top",
                                )
                            else:
                                fig.add_vrect(
                                    w.start_beat, w.end_beat,
                                    line_width=0, fillcolor=color, opacity=0.2,
                                    annotation=go.layout.Annotation(
                                        text=w.icon, yanchor="top", yref="paper",
                                        # round to inclusive 1/64, to ensure 1/192 interpolation doesn't result in ugly fractions
                                        hovertext=f"{w.note_type} {w.note_rail} @ {pretty_fraction(np.floor(w.start_beat*64)/64)} to {pretty_fraction(np.ceil(w.end_beat*64)/64)}<br>{w.text}",
                                    ),
                                    annotation_position="top",
                                )
            ui.plotly(xfig).classes("w-full h-48")
            ui.plotly(yfig).classes("w-full h-48")
            if any_vel:
                ui.plotly(vfig).classes("w-full h-48")
            if any_acc:
                ui.plotly(afig).classes("w-full h-48")

        @ui.refreshable
        def _stats_table(self) -> None:
            ui.label(f"{len(self.data.bookmarks)} Bookmarks")
            ui.label("Object counts (click to see more)")
            def _stats_notify(ev: events.GenericEventArguments) -> None:
                if "." in ev.args["colId"]:
                    col = ev.args["colId"].rsplit(".",1)[0]
                    col_data = ev.args["data"][col]
                    ui.notify(
                        f"{col_data['total']} {col}: " + pretty_list([f"{v} {k}" for k,v in col_data.items() if k != "total"]),
                        position="center",
                        type="info"
                    )
                    
            warning_counts: dict[str, dict[str, int]|int] = {
                d: -1 
                for d in self.data.difficulties.keys()
            }
            if self.warnings is not None:
                for d in self.data.difficulties.keys():
                    warnings = self.warnings.get(d, [])
                    counts: dict[str, int] = {}
                    for w in warnings:
                        counts[w.type] = counts.get(w.type, 0) + 1
                    counts["total"] = sum(counts.values())
                    warning_counts[d] = counts
            ui.aggrid({
                "domLayout": "autoHeight",
                "columnDefs": [
                    {"headerName": "Difficulty", "field": "diff"},
                    {"headerName": "Fixed Errors", "field": "errors"},
                    {"headerName": "Warnings", "field": "warnings.total"},
                    {"headerName": "Notes", "field": "notes.total"},
                    {"headerName": "Rails", "field": "rails.total"},
                    {"headerName": "Rail nodes", "field": "rail_nodes.total"},
                    {"headerName": "Walls", "field": "walls.total"},
                    {"headerName": "Lights", "field": "lights"},
                    {"headerName": "Effects", "field": "effects"},
                ],
                "rowData": [
                    c.get_counts() | {
                        "diff": d,
                        "errors": len(self.data.errors.get(d, [])),
                        "warnings": warning_counts[d]
                    }
                    for d, c in self.data.difficulties.items()
                ],
            }).classes("w-full h-auto").on("cellClicked", _stats_notify)

        @ui.refreshable
        def _hands_card(self) -> None:
            difficulty = self.storage.get("fileutils_hdiff")
            if difficulty is None:
                return
            elif self.data is None or self.hand_curves is None or self.warnings is None:
                ui.spinner(size="xl")
            elif (
                (difficulty not in self.hand_curves or all(np.isnan(pos).all() for pos, _, _ in self.hand_curves[difficulty].values()))
                and (difficulty not in self.warnings or not self.warnings[difficulty])
            ):
                ui.label("No data").classes("h-32")
            else:
                ui.label("Hand curves and warnings")
                self._hcurve_content(self.hand_curves.get(difficulty), self.warnings.get(difficulty), self.data.difficulties.get(difficulty))

        @ui.refreshable
        def _warnings_card(self) -> None:
            difficulty = self.storage.get("fileutils_wdiff")
            warning_types = self.storage.get("fileutils_warnings_types")
            note_types = self.storage.get("fileutils_warnings_notetypes")
            if difficulty is None:
                return
            elif self.data is None or self.warnings is None:
                ui.spinner(size="xl")
            elif (difficulty not in self.warnings or not self.warnings[difficulty]):
                ui.label("No data").classes("h-32")
            else:
                ui.table(
                    columns=[
                        {"name": "icon", "label": "Icon", "field": "icon", "align": "left", "sortable": True},
                        {"name": "start", "label": "Start", "field": "start", "align": "left", "sortable": True},
                        {"name": "start_pretty", "label": "Start", "field": "start_pretty", "align": "left"},
                        {"name": "end", "label": "End", "field": "end", "align": "left"},
                        {"name": "obj_type", "label": "Type", "field": "obj_type", "align": "left", "sortable": True},
                        {"name": "axis", "label": "Axis", "field": "axis", "align": "left", "sortable": True},
                        {"name": "message", "label": "Warning", "field": "message", "align": "left"},
                    ],
                    rows=[
                        {
                            "icon": w.icon.replace("<br>", " "),
                            "start": round(w.start_beat, 3),
                            "start_pretty": pretty_fraction(np.floor(w.start_beat*64)/64),
                            "end": pretty_fraction(np.ceil(w.end_beat*64)/64),
                            "obj_type": f"{w.note_type} {w.note_rail}",
                            "axis": w.figure,
                            "message": w.text.replace("<br>", " "),
                        }
                        for w in self.warnings[difficulty]
                        if w.type in warning_types
                        if w.note_type in note_types
                    ],
                    row_key="start",
                    pagination=25,
                )

        @ui.refreshable
        def _density_card(self) -> None:
            difficulty = self.storage.get("fileutils_ddiff")
            if difficulty is None:
                return
            ui.label("Wall density")
            if self.wall_densities is None:
                ui.spinner(size="xl")
            elif difficulty not in self.wall_densities or not self.wall_densities[difficulty]["combined"].max_value:
                ui.label("No data").classes("h-32")
            else:
                self._wden_content(self.wall_densities[difficulty])

            ui.label("Note & Rail density")
            if self.note_densities is None:
                ui.spinner(size="xl")
            elif difficulty not in self.note_densities or not any(pdc.max_value for pdc in self.note_densities[difficulty]["combined"].values()):
                ui.label("No data").classes("h-32")
            else:
                self._nden_content(self.note_densities[difficulty])

            ui.label("Running counts")
            if self.running_counts is None:
                ui.spinner(size="xl")
            elif difficulty not in self.running_counts or not (self.running_counts[difficulty]["notes+rails"].max_value or self.running_counts[difficulty]["notes"].max_value):
                ui.label("No data").classes("h-32")
            else:
                self._rcount_content(self.running_counts[difficulty])

        @ui.refreshable
        def stats_card(self) -> None:
            if self.data is None:
                ui.label("Load a map to show stats and graphs")
                return
            if self.merged_filenames:
                ui.label("Merged:")
                for m in self.merged_filenames:
                    ui.label(m)
            if self.data.errors:
                with ui.button("Save error report", icon="summarize", color="warning", on_click=self.save_errors):
                    ui.tooltip("Use this if you want to re-add notes that were corrupted.")
            self._stats_table()

            ui.separator()

            with ui.tabs() as tabs:
                ui.tab("bpm", label="BPM", icon="speed")
                ui.tab("hands", label="Hands", icon="accessibility")
                ui.tab("warnings", label="Warnings", icon="warning")
                ui.tab("density", label="Density", icon="analytics")

            with ui.tab_panels(tabs, value="bpm").bind_value(app.storage.user, "fileutils_stats_type").classes("w-full"):
                with ui.tab_panel("bpm"):
                    with ui.element().classes("w-full min-h-screen"):
                        self._bpm_card()
                with ui.tab_panel("hands") as hpanel:
                    hsel = ui.select({None:"select difficulty"}|{d:d for d in synth_format.DIFFICULTIES if d in self.data.difficulties}, label="Difficulty").bind_value(app.storage.user, "fileutils_hdiff").classes("w-40")
                    with ui.element().classes("w-full min-h-screen") as elem:
                        self._hands_card()
                    @handle_errors
                    async def _change_hdiff(vce: events.ValueChangeEventArguments):
                        await run.io_bound(_in_slot(self._hands_card.refresh, elem))
                    hsel.on_value_change(_change_hdiff)
                with ui.tab_panel("warnings") as hpanel:
                    with ui.row():
                        wsel = ui.select({None:"select difficulty"}|{d:d for d in synth_format.DIFFICULTIES if d in self.data.difficulties}, label="Difficulty").bind_value(app.storage.user, "fileutils_wdiff").classes("w-40")
                        ntypesel = ui.select(
                            list(synth_format.NOTE_TYPES),
                            value=list(synth_format.NOTE_TYPES),
                            label="Note types",
                            multiple=True,
                        ).bind_value(app.storage.user, "fileutils_warnings_notetypes").classes("w-48")
                        wtypesel = ui.select(
                            list(analysis.WARNING_TYPES),
                            value=list(analysis.WARNING_TYPES),
                            label="Warning types",
                            multiple=True,
                        ).bind_value(app.storage.user, "fileutils_warnings_types").classes("w-auto")
                    with ui.element().classes("w-full min-h-screen") as elem:
                        self._warnings_card()
                    @handle_errors
                    async def _change_w(vce: events.ValueChangeEventArguments):
                        await run.io_bound(_in_slot(self._warnings_card.refresh, elem))
                    wsel.on_value_change(_change_w)
                    wtypesel.on_value_change(_change_w)
                    ntypesel.on_value_change(_change_w)
                with ui.tab_panel("density") as dpanel:
                    dsel = ui.select({None:"select difficulty"}|{d:d for d in synth_format.DIFFICULTIES if d in self.data.difficulties}, label="Difficulty").bind_value(app.storage.user, "fileutils_ddiff").classes("w-40")
                    with ui.element().classes("w-full min-h-screen") as elem:
                        self._density_card()
                    @handle_errors
                    async def _change_ddiff(vce: events.ValueChangeEventArguments):
                        await run.io_bound(_in_slot(self._density_card.refresh, elem))
                    dsel.on_value_change(_change_ddiff)
        def __repr__(self) -> str:
            return type(self).__name__  # avoid spamming logs with binary data

    fi = FileInfo()

    with ui.dialog() as help_dialog, ui.card():
        ui.markdown("""
            **This tab allows you to work .synth files directly.**

            The following features are supported:

            * Create new .synth files from audio files
                * many common formats (including `.mp3`) will be converted to `.ogg` via [libsndfile](http://www.mega-nerd.com/libsndfile)
            * Detect and edit BPM/Offset (without changing timing of existing objects)
                * Advanced audio processing to detect sections with different BPM via [librosa](https://librosa.org/)
                * Graphical representation of intermediate data (via [plotly](https://plotly.com/python/))
                * Manual BPM override
                * Shift offset by half beats for easy alignment
            * Replace audio file
            * View and edit metadata:
                * name, artist and mapper
                * cover image
            * Detect and correct certain types of file corruption or errors (NaN values, duplicate notes)
            * Merge files, including different BPM and Offset
            * Show stats
                * Object counts per difficulty
                * Hand position, velocity and acceleration plots
                * Density plot for Walls (including checks for PC or Quest limitations)
                * Density plot for Notes and Rails
            * See warnings about problematic sections (spiral distortions, headbanger notes)

            To start, just open a .synth file by clicking the plus button below.  
            You can also drag files directly onto these file selectors.

            Note: Sometimes the file upload gets stuck. In that case just press the button twice more (first time it will be an "X", the second time a upload-cloud icon)
        """)
    ui.button("What can I do here?", icon="help", color="info", on_click=help_dialog.open)
    
    with ui.row().classes("mb-4"):
        with ui.card().classes("mb-4 w-72"):
            ui.markdown("**Base Map**")
            ui.separator()
            with ui.upload(label="Select a .synth or audio file ->", auto_upload=True, on_upload=fi.upload).props('accept=".synth,audio/*,*"').classes("w-full").add_slot("list"):
                ui.tooltip("Select a file first").bind_visibility_from(fi, "is_valid", backward=lambda v: not v).classes("bg-red")
                ui.input("Output Filename").props("dense").bind_value(fi, "output_filename").bind_enabled_from(fi, "is_valid")
                with ui.switch("Finalize Walls").bind_value(fi, "output_finalize").bind_enabled_from(fi, "is_valid").classes("my-auto"):
                    ui.tooltip("Shifts some walls down, such that they look ingame as they do in the editor")
                with ui.row():
                    ui.button("clear", icon="clear", color="negative", on_click=fi.clear).props("dense").classes("w-28").bind_enabled_from(fi, "is_valid")
                    ui.button("save", icon="save", color="positive", on_click=fi.save).props("dense").classes("w-28 ml-auto").bind_enabled_from(fi, "is_valid")
        with ui.card().classes("w-100").bind_visibility(fi, "is_valid"):
            fi.info_card()
        with ui.card().bind_visibility(fi, "is_valid"):
            ui.markdown("**Merge files into base**")
            ui.separator()
            merge_bookmarks = ui.switch("Merge Bookmarks", value=True).classes("w-full").bind_value(app.storage.user, "merge_bookmarks").tooltip("Disable this if you merge maps that contain the same bookmarks")
            with ui.upload(label="One or more files", multiple=True, auto_upload=True, on_upload=fi.upload_merge).props('color="positive" accept=".synth,*"').classes("w-full").add_slot("list"):
                ui.markdown("""
                    Should have the same audio file as the base.  
                    BPM & Offset will be matched automatically.
                """)
            

    with ui.card().classes("w-full"):
        fi.stats_card()

file_utils_tab = GUITab(
    name="fileutils",
    label="File Utils",
    icon="construction",
    content_func=_file_utils_tab,
)



import asyncio
from contextlib import contextmanager
from concurrent.futures.process import BrokenProcessPool
from dataclasses import dataclass
import datetime
from functools import wraps
from io import BytesIO
import json
import math
import logging
import traceback
from typing import Any, Callable, Generator, Optional

from fastapi import Response
from nicegui import app, events, ui, binding, run
from nicegui.storage import PersistentDict
import pyperclip

from .. import synth_format, utils, __version__

__all__ = [
    "logger", "wiki_base",
    "GUITab", "SMHInput",
    "error", "warning", "info",
    "wiki_reference", "try_load_synth_file", "add_suffix",
    "ParseInputError", "PrettyError", "handle_errors",
    "read_clipboard", "write_clipboard", "safe_clipboard_data",
]

logger = logging.getLogger("SMH-GUI")
wiki_base = "https://github.com/adosikas/synth_mapping_helper/wiki"

start_time = datetime.datetime.now(datetime.timezone.utc)
last_errors: list[tuple[str, Exception|None, Any, Any, datetime.datetime]] = []

@dataclass
class GUITab:
    name: str
    label: str
    icon: str
    content_func: Callable

    def get_settings(self) -> dict[str, Any]:
        return {
            s_id.removeprefix(self.name + "_"): v
            for s_id, v in app.storage.user.items()
            if s_id.startswith(self.name)
        }

    def delete_settings(self) -> None:
        count = 0
        for s_id in list(app.storage.user):
            if s_id.startswith(self.name):
                count += 1
                del app.storage.user[s_id]
        logger.info(f"Deleted {count} settings for {self.label} tab")
        ui.navigate.reload()


class SMHInput(ui.input):
    def __init__(self,
        storage_id: Optional[str],
        label: str,
        default_value: str|float,
        allow_empty: bool = False,
        tooltip: Optional[str] = None,
        suffix: Optional[str] = None,
        negate_icons: dict[int, str]|None = None,
        tab_id: Optional[str] = None,
        width: int = 12,
        height: int = 10,
        **input_kwargs: Any,
    ):
        self.allow_empty = allow_empty
        super().__init__(label=label, value=str(default_value), validation=self._validate, **input_kwargs)
        if storage_id is not None:
            if tab_id is not None:
                storage_id = f"{tab_id}_{storage_id}"
            self.bind_value(app.storage.user, storage_id)
        self.classes(f"w-{width} h-{height}")
        self.props('dense input-style="text-align: right" no-error-icon')
        self.storage_id = storage_id
        self.default_value = default_value
        if suffix:
            self.props(f'suffix="{suffix}"')
        if tooltip is not None:
            self.tooltip(tooltip)
        if negate_icons is not None:
            negate_icons = {1: "add", 0: "close", -1: "remove"} | negate_icons
            def _negate(val: str|None) -> str|None:
                if not val:
                    return val
                if val.startswith("-"):
                    return val[1:]
                return "-" + val
            def _get_icon(val: str|None) -> str:
                if not val:  # empty string or None
                    return negate_icons[0]
                try:
                    v = utils.parse_number(val)
                except ValueError:
                    return "error"
                if v > 0:
                    return negate_icons[1]
                if v < 0:
                    return negate_icons[-1]
                return negate_icons[0]
            with self.add_slot("prepend"):
                self.icon = ui.icon("", color="primary").classes("border-2 rounded cursor-pointer").on(
                    "click", lambda e: self.set_value(_negate(self.value))
                ).bind_name_from(self, "value", _get_icon)
                ui.tooltip("Click to negate")
        with self.add_slot("error"):
            ui.element().style("visiblity: hidden")

    def _validate(self, value: Any) -> None|str:
        if self.allow_empty and not value:
            return None
        try:
            v = utils.parse_number(value)
            return None
        except ValueError:
            return ""

    @property
    def parsed_value(self) -> float | None:
        if self.allow_empty and not self.value:
            return None
        try:
            return utils.parse_number(self.value)
        except ValueError as ve:
            raise ParseInputError(input_id=self.storage_id or "input", value=self.value, exc=ve) from ve

class PrettyJSONResponse(Response):
    media_type = "application/json"

    def render(self, content: Any) -> bytes:
        return json.dumps(content, indent=2).encode("utf-8")

@app.get("/error_report", response_class=PrettyJSONResponse)
def error_report():
    now = datetime.datetime.now(datetime.timezone.utc)
    out = {
        "version": __version__,
        "report_time": now.strftime("%Y-%m-%dT%H-%M-%SZ"),
        "runtime": str(datetime.timedelta(seconds=round((now-start_time).total_seconds()))),  # rounded to full seconds
        "errors": [],
    }
    for msg, exc, context, data, time in last_errors:
        err = {
            "time": time.strftime("%Y-%m-%dT%H-%M-%SZ"),
            "message": msg,
        }
        if exc is not None:
            err["exception"] = repr(exc)
            if exc.__cause__ is not None:
                err["exception_cause"] = repr(exc.__cause__)
            err["traceback"] = [l for tb in traceback.format_exception(exc, chain=True) for l in tb.splitlines()]
        if context is not None:
            err["context"] = context
        if data is not None:
            err["data"] = data
        out["errors"].append(err)
    return out

def error(msg: str, exc: Optional[Exception]=None, context: Any = None, data: Any=None) -> None:
    last_errors.append((msg, exc, context, data, datetime.datetime.now(datetime.timezone.utc)))
    logger.error(msg+(" " + repr(exc) if exc is not None else ""))
    if exc is not None:
        logger.debug("Stacktrace:", exc_info=exc)
    ui.notify(msg, type="negative", progress=True, group=False, caption=repr(exc) if exc is not None else None)

def warning(msg: str) -> None:
    logger.warning(msg)
    ui.notify(msg, type="warning", progress=True, group=False)

def info(msg: str, caption: Optional[str]=None) -> None:
    logger.info(msg)
    ui.notify(msg, type="positive", progress=True, caption=caption)

def wiki_reference(page: str, invert_colors: bool = False) -> ui.badge:
    b = ui.badge("?").style("cursor: help")
    if invert_colors:
        b.props("color=white text-color=primary")
    # .stop to stop event bubbling up to containers (ie button or switch)
    b.on("click.stop", lambda _ : ui.navigate.to(f"{wiki_base}/{page}", new_tab=True))
    with b:
        ui.tooltip(f"Open wiki: {page}")
    return b

def try_load_synth_file(e: events.UploadEventArguments) -> Optional[synth_format.SynthFile]:
    try:
        data = synth_format.import_file(BytesIO(e.content.read()))
    except Exception as exc:
        msg = f"Error reading {e.name} as SynthFile"
        upl: ui.upload = e.sender  # type: ignore
        upl.reset()
        error(msg, exc)
        return None
    if data.errors:
        warning("There was invalid data in the file, affected items were REMOVED:")
        for diff, errors in data.errors.items():
            if len(errors) > 3:
                warning(f"{diff}: {len(errors)} errors. See the error report for details.")
            else:
                for jpe, time in errors:
                    warning(f"{diff}@{time}: {jpe}")
    return data

def add_suffix(filename: str, suffix: str) -> str:
    if not suffix:
        return filename
    return f"{filename.removesuffix('.synth')}_{suffix}.synth"

class ParseInputError(ValueError):
    def __init__(self, input_id: str, value: Any, exc: Optional[ValueError]=None, context: Any = None) -> None:
        self.input_id = input_id
        self.value = value
        self.exc = exc
        self.context = context

class PrettyError(RuntimeError):
    def __init__(self, msg: str, exc: Optional[Exception]=None, context: Any = None, data: Any = None) -> None:
        self.msg = msg
        self.exc = exc
        self.context = context
        self.data = data

# decorator for pretty error handling
def handle_errors(func: Callable) -> Callable:
    # Usage:
    #   @handle_errors
    #   def _do_stuff(...):
    #     try:
    #       ...
    #     except ... as ...:
    #       raise PrettyError(...)
    #   ui.button(on_click=_do_stuff)
    # Lambdas:
    #   ui.button(on_click=handle_errors(lambda ...))
    @contextmanager
    def _catch_errors():
        try:
            yield
        except ParseInputError as pie:
            error(msg=f"Error parsing {pie.input_id} value: {pie.value!r}", exc=pie.exc, context=pie.context, data=pie.value)
        except PrettyError as pe:
            error(msg=pe.msg, exc=pe.exc, context=pe.context, data=pe.data)
        except BrokenProcessPool as bpe:
            error(msg="Sound library error occurred. I am waiting for a fix for that. In the meantime, avoid this function.", exc=bpe)
            run.process_pool = run.ProcessPoolExecutor()  # attempt to restore process pool
        except Exception as exc:
            error(msg="Unexpected error. Please report this by clicking the report button in the top-right and sending me the report.", exc=exc)

    if asyncio.iscoroutinefunction(func):
        @wraps(func)
        async def _wrapped_async_func(*args, **kwargs) -> Any:
            with _catch_errors():
                return await func(*args, **kwargs)
        return _wrapped_async_func
    @wraps(func)
    def _wrapped_sync_func(*args, **kwargs) -> Any:
        with _catch_errors():
            return func(*args, **kwargs)
    return _wrapped_sync_func

# wrappers, for when I decide to use the browser clipboard
def read_clipboard() -> str:
    return pyperclip.paste()

def write_clipboard(text: str) -> None:
    pyperclip.copy(text)

# like synth_format.clipboard_data, but reporting errors
@contextmanager
def safe_clipboard_data(use_original: bool = False, realign_start: bool = True, write: bool = True) -> Generator[synth_format.ClipboardDataContainer, None, None]:
    try:
        clipboard_in = read_clipboard()
    except RuntimeError as re:
        raise PrettyError(msg=f"Error reading clipboard", exc=re)
    if not clipboard_in:
        raise PrettyError(msg="Clipboard does not contain any data")
    if not clipboard_in.startswith("{") or not clipboard_in.endswith("}"):
        raise PrettyError(msg="Clipboard does not contain JSON data")
    try:
        data = synth_format.ClipboardDataContainer.from_json(clipboard_in, use_original=use_original)
    except (KeyError, ValueError) as fje:
        raise PrettyError(msg="Error reading data in clipboard", exc=fje, data=clipboard_in)

    # don't catch any errors here, so clipboard is not written to on error
    yield data

    if not write:
        return

    try:
        clipboard_out = data.to_clipboard_json(realign_start=realign_start)
    except ValueError as ve:
        raise PrettyError(msg=f"Error outputting data for clipboard", exc=ve)
    try:
        write_clipboard(clipboard_out)
    except RuntimeError as re:
        raise PrettyError(msg=f"Error writing clipboard", exc=re)

class PreventDefaultKeyboard(ui.keyboard, component="prevent_default_keyboard.js"):
    def bind_active_from(self,
                        target_object: Any,
                        target_name: str = 'active',
                        backward: Callable[..., Any] = lambda x: x,
                        ) -> "Self":
        binding.bind_from(self, 'active', target_object, target_name, backward)
        return self



from functools import wraps
import numpy as np

from .synth_format import WALL_MIRROR_ID, WALL_TYPES

# Note: None of these functions are allowed to *modify* the input array instance. Returning the same array (if nothing needed to be changed) is allowed.

MIRROR_VEC_3D = np.array([-1, 1, 1])

def add_basic_pivot_wrapper(func):
    @wraps(func)
    def _pivot_wrapper(
        data: "numpy array (n, 3+)",
        *args,
        relative: bool = False,
        pivot: "optional numpy array (2+)"=None,
        **kwargs
    ) ->  "numpy array (n, 3+)":
        if relative:
            pivot = data[0,:3]
        if pivot is not None and pivot.any():
            pivot_nd = np.zeros(data.shape[-1])
            pivot_nd[:pivot.shape[0]] = pivot
            return func(data-pivot_nd, *args, **kwargs) + pivot_nd
        return func(data, *args, **kwargs)
    return _pivot_wrapper

# movement
def _offset(data: "numpy array (n, m)", offset_3d: "numpy array (3)", direction: int = 1) -> "numpy array (n, 3+)":
    """translate positions"""
    offset_nd = np.zeros((data.shape[-1]))
    offset_nd[..., :3] = offset_3d if direction == 1 else offset_3d * MIRROR_VEC_3D
    return data + offset_nd


def _offset_relative(
    data: "numpy array (n, m)", offset_3d: "numpy array (3)", direction: int = 1
) -> "numpy array (n, 3+)":
    """translate positions, use relative coordinates for walls"""
    if data.shape[-1] == 3:
        return _offset(data, offset_3d, direction=direction)
    # calculate rot matrix for angle of every wall
    rad_ang = np.radians(np.atleast_1d(data[:, 4]))
    rot_matrix = np.rollaxis(
        np.array(
            (
                (np.cos(rad_ang), np.sin(rad_ang)),
                (-np.sin(rad_ang), np.cos(rad_ang)),
            )
        ),
        -1,
    )
    offset_nd = np.zeros((data.shape[-1]))
    offset_nd[..., :2] = (
        np.array(offset_3d[:2]) if direction == 1 else offset_3d[:2] * MIRROR_VEC_3D
    ).dot(
        rot_matrix
    )  # offset is rotated for each wall
    offset_nd[..., 2] = offset_3d[2]  # t stays as-is
    return data + offset_nd

def offset(
    data: "numpy array (n, 3+)",
    offset_3d: "numpy array (3)",
    pivot: None = None,  # ignored, for compatibility with other movement funcs
    relative: bool = False,
    direction: int = 1,
) -> "numpy array (n, 3+)":
    if relative:
        return _offset_relative(data, offset_3d=offset_3d, direction=direction)
    else:
        return _offset(data, offset_3d=offset_3d, direction=direction)

@add_basic_pivot_wrapper
def outset(
    data: "numpy array (n, m)", outset_scalar: float, direction: int = 1
) -> "numpy array (n, 3+)":
    """move positions outwards"""
    zero_mask = np.logical_and(np.isclose(data[..., 0], 0, atol=1e-5), np.isclose(data[..., 1], 0, atol=1e-5))  # ignore xy close to 0,0
    angles = np.arctan2(data[~zero_mask, 1], data[~zero_mask, 0])
    normalized = np.zeros(data.shape)
    normalized[~zero_mask, 0] = np.cos(angles)
    normalized[~zero_mask, 1] = np.sin(angles)
    return data + normalized * outset_scalar

@add_basic_pivot_wrapper
def scale(
    data: "numpy array (n, 3+)", scale_3d: "numpy array (3)", direction: int = 1
) -> "numpy array (n, 3+)":
    """scale positions relative to center and start of selection"""
    if scale_3d[2] == 0:
        raise ValueError("Cannot have 0 for time scale")
    scale_nd = np.ones((data.shape[-1]))
    scale_nd[..., :3] = scale_3d
    output = data * scale_nd
    if scale_nd[2] < 0:  # reverse order of elements
        output = output[::-1]
    if data.shape[-1] == 5: # walls
        if (scale_nd[0] < 0) != (scale_nd[1] < 0):  # mirror X *or* Y: swap type and invert angle (both: do nothing)
            output[:,3] = [WALL_MIRROR_ID[i] for i in output[:,3]]
            output[:,4] = -output[:,4]
        if (scale_nd[1] < 0):  # mirror Y: add 180
            output[:,4] += 180

    return output


@add_basic_pivot_wrapper
def rotate(
    data: "numpy array (n, 3+)", angle: float, direction: int = 1
) -> "numpy array (n, 3+)":
    """rotate positions anticlockwise around center"""
    rad_ang = np.radians(angle * direction)
    rot_matrix = np.identity(data.shape[-1])
    rot_matrix[:2, :2] = [
        [np.cos(rad_ang), np.sin(rad_ang)],
        [-np.sin(rad_ang), np.cos(rad_ang)],
    ]
    out = data.dot(rot_matrix)
    if data.shape[-1] >= 5:
        # just add to wall rotation (unless it is a crouch wall)
        not_crouch = (out[..., 3] != WALL_TYPES["crouch"][0])
        out[not_crouch, 4] += angle * direction
    return out



from typing import Literal

import numpy as np
from scipy.interpolate import CubicHermiteSpline, pchip_interpolate

from .synth_format import DataContainer, SINGLE_COLOR_NOTES
from .utils import bounded_arange_plusminus

# How far last node and start note can be spaced for two rails to be merged
MERGE_ACCURACY_GRID = 1 / 8
MERGE_ACCURACY_BEAT = 1 / 64

def interpolate_linear(data: "numpy array (n, m)", new_z: "numpy array (x)", *, direction: int = 1) -> "numpy array (x, 3)":
    return np.stack((
        np.interp(new_z, data[:, 2], data[:, 0]),
        np.interp(new_z, data[:, 2], data[:, 1]),
        new_z,
    ), axis=-1)

def interpolate_hermite(data: "numpy array (n, m)", new_z: "numpy array (x)", *, direction: int = 1) -> "numpy array (x, 3)":
    return np.stack((
        pchip_interpolate(data[:, 2], data[:, 0], new_z),
        pchip_interpolate(data[:, 2], data[:, 1], new_z),
        new_z,
    ), axis=-1)

def synth_curve(data: "numpy array (n, m)", *, direction: int = 1):
    # based on https://github.com/LittleAsi/synth-riders-editor,
    # which uses LineSmoother.cs from https://forum.unity.com/threads/easy-curved-line-renderer-free-utility.391219
    if data.shape[0] == 1:
        return data

    # equivalent of Unity's AnimationCurve when doing smoothTangents(0)
    tangents = np.diff(data, axis=0, prepend=data[0][np.newaxis], append=data[-1][np.newaxis])
    smoothed_tangents = (tangents[:-1] + tangents[1:]) / 2

    # this interpolates not only x and y, but also time
    curve = CubicHermiteSpline(range(data.shape[0]), data, smoothed_tangents)
    curve_points = []
    for i in range(data.shape[0]-1):
        curve_points.append(data[i])
        # interpolate a number of points which is based on distance between nodes
        # intermediates = int(np.linalg.norm(coord_to_synth(240, data[i+1]-data[i])) / 0.15)

        # we use a different formula, which doesn't rely on BPM but produces very similar results
        intermediates = int(np.linalg.norm((data[i+1]-data[i])*(0.1,0.1,16)))
        curve_points.extend(curve(i + np.arange(1, intermediates)/intermediates))
    curve_points.append(data[-1])
    return np.array(curve_points)

def interpolate_spline(data: "numpy array (n, m)", new_z: "numpy array (x)", *, direction: int = 1) -> "numpy array (x, 3)":
    return interpolate_linear(synth_curve(data), new_z, direction=direction)

def get_position_at(notes: SINGLE_COLOR_NOTES, beat: float, interpolate_gaps: bool = True) -> "numpy array (2)":
    # single note
    if beat in notes:
        return notes[beat][0, 0:2].copy()
    last_before = None
    # on rail, or between notes: interpolate
    for time in sorted(notes):
        nodes = notes[time]
        if time > beat:  # passed by target beat: interpolate between last and this
            if interpolate_gaps and last_before is not None:
                return interpolate_spline(np.concatenate((last_before, nodes)), beat)[0:2]
            break
        last_before = nodes
        if nodes.shape[0] == 1:  # ignore single nodes
            continue
        if nodes[-1, 2] >= beat:  # on rail: interpolate between nodes
            return interpolate_spline(nodes, beat)[0:2]

    # neither
    return None

# Note: None of these functions are allowed to *modify* the input dict instance. Returning the same dict (if nothing needed to be changed) is allowed.

def split_rails(notes: SINGLE_COLOR_NOTES, *, direction: int = 1) -> SINGLE_COLOR_NOTES:
    """splits rails at single notes"""
    current_rail_start = None
    current_rail_end = None
    out: SINGLE_COLOR_NOTES = {}

    for time in sorted(notes):
        if current_rail_start is not None and time >= current_rail_end:
            current_rail_start = None
            current_rail_end = None
        nodes = notes[time]
        if nodes.shape[0] > 1:
            current_rail_start = nodes[0, 2]
            current_rail_end = nodes[-1, 2]
            out[time] = nodes
        elif current_rail_start is not None:
            previous_nodes = out[current_rail_start]
            # last index that is *before* or *at* the note
            last_index = np.argwhere(previous_nodes[:, 2] <= nodes[0, 2])[-1][0]

            # use the remainder of current rail for new rail
            # includes [last_index] that gets replaced by note to avoid concatenating
            remainder = previous_nodes[last_index:].copy()
            remainder[0] = nodes[0]
            out[time] = remainder

            if previous_nodes[last_index, 2] == time:
                # when there is a existing node at the current time, leave it be
                out[current_rail_start] = previous_nodes[: last_index + 1]
            else:
                # else replace the next one with note to finish the rail and include it
                previous_nodes[last_index + 1] = nodes[0]
                out[current_rail_start] = previous_nodes[: last_index + 2]

            current_rail_start = remainder[0, 2]
            current_rail_end = remainder[-1, 2]
        else:
            # ignore freestanding single note
            out[time] = nodes
    return out

def snap_singles_to_rail(notes: SINGLE_COLOR_NOTES, *, direction: int = 1) -> SINGLE_COLOR_NOTES:
    """snap single notes to rail"""
    current_rail_start = None
    current_rail_end = None
    out: SINGLE_COLOR_NOTES = {}

    for time in sorted(notes):
        if current_rail_start is not None and time > current_rail_end:  # current time is past end of stored rail -> forget stored rail
            current_rail_start = None
            current_rail_end = None
        nodes = notes[time]
        if nodes.shape[0] > 1:  # rail: store start/end and keep as-is
            current_rail_start = nodes[0, 2]
            current_rail_end = nodes[-1, 2]
            out[time] = nodes
        elif current_rail_start is not None:  # single next to rail -> snap to it
            out[time] = interpolate_spline(out[current_rail_start], [time])
        else:  # single without rail next to it: keep as-is
            out[time] = nodes
    return out

def merge_sequential_rails(notes: SINGLE_COLOR_NOTES, *, direction: int = 1) -> SINGLE_COLOR_NOTES:
    """merges rails where end and start a very close"""
    current_rail_start = None
    current_rail_end = None
    out: SINGLE_COLOR_NOTES = {}

    for time in sorted(notes):
        nodes = notes[time]
        if nodes.shape[0] == 1:  # ignore single nodes
            out[time] = nodes
            continue

        if (
            current_rail_end is not None
            and time <= current_rail_end + MERGE_ACCURACY_BEAT
        ):
            previous_nodes = out[current_rail_start]
            if np.isclose(
                previous_nodes[-1, 2], nodes[0, 2], atol=MERGE_ACCURACY_BEAT
            ) and np.allclose(
                previous_nodes[-1, :2], nodes[0, :2], atol=MERGE_ACCURACY_GRID
            ):
                # add to rail
                out[current_rail_start] = np.concatenate((previous_nodes[:-1], nodes))
                current_rail_end = nodes[-1, 2]
                # add single note where rail started
                out[time] = nodes[:1]
                continue
    
        current_rail_start = time
        current_rail_end = nodes[-1, 2]
        out[time] = nodes

    return out

def merge_rails(notes: SINGLE_COLOR_NOTES, max_interval: float, *, direction: int = 1) -> SINGLE_COLOR_NOTES:
    """merges rails"""
    current_rail_start = None
    current_rail_end = None
    out: SINGLE_COLOR_NOTES = {}

    for time in sorted(notes):
        nodes = notes[time]
        if nodes.shape[0] == 1:  # ignore single nodes
            out[time] = nodes
            continue

        if (
            current_rail_end is None
            or time > current_rail_end + max_interval
        ):
            current_rail_start = time
            current_rail_end = nodes[-1, 2]
            out[time] = nodes
        else:
            previous_nodes = out[current_rail_start]
            # when start and end are close, remove end node of previous rail before joining
            if np.isclose(
                previous_nodes[-1, 2], nodes[0, 2], atol=MERGE_ACCURACY_BEAT
            ) and np.allclose(
                previous_nodes[-1, :2], nodes[0, :2], atol=MERGE_ACCURACY_GRID
            ):
                previous_nodes = previous_nodes[:-1]

            # add to rail
            out[current_rail_start] = np.concatenate((previous_nodes, nodes))
            current_rail_end = nodes[-1, 2]
            # add single note where rail started
            out[time] = nodes[:1]

    return out

def connect_singles(notes: SINGLE_COLOR_NOTES, max_interval: float, *, direction: int = 1) -> SINGLE_COLOR_NOTES:
    """Turn single notes into rails"""
    current_rail_start = None
    current_rail_end = None
    out: SINGLE_COLOR_NOTES = {}
    for time in sorted(notes):
        nodes = notes[time]

        if nodes.shape[0] != 1:  # existing rails are added unchanged and stops notes from connecting past it
            current_rail_start = None
            current_rail_end = None
            out[time] = nodes
            continue

        if (
            current_rail_end is None
            or time > current_rail_end + max_interval + MERGE_ACCURACY_BEAT
        ):
            # start a new rail (but only 1 node for now)
            current_rail_start = time
            current_rail_end = nodes[-1, 2]
            out[time] = nodes
        else:
            # extend the current rail inplace
            out[current_rail_start] = np.concatenate((out[current_rail_start], nodes))
            current_rail_end = nodes[-1, 2]
            # note: does not readd the single notes to the output dict

    return out

def interpolate_nodes(
    data: "numpy array (n, 3)", mode: Literal["spline", "hermite", "linear"], interval: float, *,
    direction: int = 1, relative: bool = False, pivot: "optional numpy array (2+)"=None,
) -> "numpy array (n, 3)":
    """places nodes at defined interval along the rail, interpolating between existing nodes. Can be negative to start from end."""
    if data.shape[0] == 1:  # ignore single nodes
        return data
    new_z = bounded_arange_plusminus(data[0, 2], data[-1, 2], interval)

    if mode == "spline":
        return interpolate_spline(data, new_z)
    elif mode == "hermite":
        return interpolate_hermite(data, new_z)
    elif mode == "linear":
        return interpolate_linear(data, new_z)
    else:
        raise RuntimeError("Invalid iterpolation mode")

def rails_to_singles(notes: SINGLE_COLOR_NOTES, keep_rail: bool, *, direction: int = 1) -> SINGLE_COLOR_NOTES:
    """Turn all rails into single notes"""
    out: SINGLE_COLOR_NOTES = {}
    for time in sorted(notes):
        nodes = notes[time]
        if nodes.shape[0] == 1:  # ignore single nodes
            out[time] = nodes
            continue
        if keep_rail:
            out[nodes[0,2]] = nodes
        for node in nodes[int(keep_rail):]:  # when keeping the rail, don't overwrite the start with a single note
            out[node[2]] = node[np.newaxis]

    return out

def rails_to_notestacks(notes: SINGLE_COLOR_NOTES, interval: float, keep_rail: bool, *, direction: int = 1) -> SINGLE_COLOR_NOTES:
    """Turn all rails into notestacks. Can be negative to start from wnd"""
    out: SINGLE_COLOR_NOTES = {}
    for time in sorted(notes):
        nodes = notes[time]
        if nodes.shape[0] == 1:  # ignore single nodes
            out[time] = nodes
            continue
        i_nodes = interpolate_nodes(nodes, "spline", interval=interval)
        if keep_rail:
            out[nodes[0,2]] = nodes
        for node in i_nodes[int(keep_rail):]:  # when keeping the rail, don't overwrite the start with a single note
            out[node[2]] = node[np.newaxis]

    return out

def shorten_rail_by(
    data: "numpy array (n, 3)", distance: float, *,
    direction: int = 1, relative: bool = False, pivot: "optional numpy array (2+)"=None,
) -> "numpy array (n, 3)":
    """Cut a bit of the end or start (if negative) of the rail"""
    if data.shape[0] == 1 or not distance:  # ignore single nodes
        return data
    if distance > 0:  # cut at the end
        if (data[-1,2] - data[0,2]) <= distance:  # shorter -> return rail start only
            return data[0][np.newaxis]
        new_z = data[-1,2] - distance
        last_index = np.argwhere(data[:, 2] < new_z)[-1][0]  # last node before new end
        return np.concatenate((data[:last_index+1], interpolate_spline(data, [new_z])))
    else:  # cut at the start
        if (data[-1,2] - data[0,2]) <= -distance:  # shorter -> return rail end only
            return data[-1][np.newaxis]
        new_z = data[0,2] - distance  # distance is negative, so "- distance" is correct here
        first_index = np.argwhere(data[:, 2] > new_z)[0][0]  # first node after new start
        return np.concatenate((interpolate_spline(data, [new_z]), data[first_index:]))

def shorten_rail_to(
    data: "numpy array (n, 3)", distance: float, *,
    direction: int = 1, relative: bool = False, pivot: "optional numpy array (2+)"=None,
) -> "numpy array (n, 3)":
    """Cut the end or start (if negative) of the rail to be as long as specified"""
    if data.shape[0] == 1 or data[-1,2] - data[0,2] <= distance:  # ignore single nodes or shorter rails
        return data
    if not distance:  # distance 0 -> return rail start only
        return data[0][np.newaxis]
    elif distance > 0:  # cut at the end
        new_z = data[0,2] + distance
        last_index = np.argwhere(data[:, 2] < new_z)[-1][0]  # last node before new end
        return np.concatenate((data[:last_index+1], interpolate_spline(data, [new_z])))
    else:  # cut at the start
        new_z = data[-1,2] + distance  # distance is negative, so "+ distance" is correct here
        first_index = np.argwhere(data[:, 2] > new_z)[0][0]  # first node after new start
        return np.concatenate((interpolate_spline(data, [new_z]), data[first_index:]))

def extend_level(
    data: "numpy array (n, 3)", distance: float, *,
    direction: int = 1, relative: bool = False, pivot: "optional numpy array (2+)"=None,
) -> "numpy array (n, 3)":
    """Add a level rail section at the end or start (if negative). Turns single notes into rails"""
    if not distance:
        return data
    elif distance > 0:
        return np.concatenate((data, data[np.newaxis,-1]+[0,0,distance]))
    else:
        return np.concatenate((data[np.newaxis,0]+[0,0,distance], data))

def extend_straight(
    data: "numpy array (n, 3)", distance: float, *,
    direction: int = 1, relative: bool = False, pivot: "optional numpy array (2+)"=None,
) -> "numpy array (n, 3)":
    """Add a straight rail section at the end or start (if negative) which keeps the same direction of the previous segment"""
    if data.shape[0] == 1 or not distance:  # ignore single nodes or shorter rails
        return data
    elif distance > 0:
        delta = data[-1] - data[-2]
        return np.concatenate((data, data[np.newaxis,-1]+delta*(distance/delta[2])))
    else:
        delta = data[0] - data[1]
        return np.concatenate((data[np.newaxis,0]+delta*(distance/delta[2]), data))

def extend_to_next(
    notes: SINGLE_COLOR_NOTES, distance: float, *, direction: int = 1
) -> SINGLE_COLOR_NOTES:
    """Add a rail section at the end pointing to next note/rail. Vice-versa for negative. Turns single notes into rails."""
    if len(notes) < 2 or not distance:  # ignore when there is nothing to work with
        return notes
    out: SINGLE_COLOR_NOTES = {}
    last_nodes: "numpy array (n, 3)" = None
    if distance > 0:
        for time in sorted(notes):
            nodes = notes[time]
            if last_nodes is not None:
                delta = nodes[0] - last_nodes[-1]
                out[last_nodes[0,2]] = np.concatenate((last_nodes, last_nodes[np.newaxis,-1]+delta*(distance/delta[2])))
            last_nodes = nodes
        out[last_nodes[0,2]] = last_nodes
    else:
        for time in sorted(notes):
            nodes = notes[time]
            if last_nodes is not None:
                delta = nodes[0] - last_nodes[-1]
                new = np.concatenate((nodes[np.newaxis,0]+delta*(distance/delta[2]), nodes))
                out[new[0,2]] = new
            else:
                out[nodes[0,2]] = nodes
            last_nodes = nodes
    return out

def segment_rail(
    notes: SINGLE_COLOR_NOTES, max_length: float, *, direction: int = 1
) -> SINGLE_COLOR_NOTES:
    """Segment rails into multiple range of maximum length. Can be negative to segment from end"""
    out: SINGLE_COLOR_NOTES = {}
    for time in sorted(notes):
        nodes = notes[time]
        if nodes.shape[0] == 1 or (nodes[-1,2]-nodes[0,2]) <= max_length:  # ignore single nodes or short rails
            out[time] = nodes
            continue

        steps = bounded_arange_plusminus(nodes[0,2], nodes[-1,2], max_length)
        new_z = np.union1d(nodes[:,2], steps)
        for start in steps[:-1]:
            out[start] = interpolate_spline(nodes, new_z=new_z[np.logical_and(new_z>=start, new_z<=(start+abs(max_length)))])
    return out

def reinterpolation_smoothing(notes: SINGLE_COLOR_NOTES, iterations: int = 3, resistance: float = 1.0, singles_mode: Literal["snap", "anchor", "temp_anchor"] = "anchor", *, direction: int = 1) -> SINGLE_COLOR_NOTES:
    # The idea here is that we alternate between which nodes get reinterpolated
    # To constrain this, start and end of the rail, optionally all single notes act as anchors which the output must pass through
    # Below, SNE are the anchor nodes, and _ are the nodes that get reinterpolated during the alternating A and B steps
    # A: S_a_aN_a_E
    # B: Sb_b_Nb_bE
    if resistance < 0:
        raise ValueError("Smoothing resistance must not be negative")
    singles: SINGLE_COLOR_NOTES = {}
    # split out rails and single notes
    rails: list["numpy array (n, 3)"] = []
    for time in sorted(notes):
        if notes[time].shape[0] == 1:
            singles[time] = notes[time]
        else:
            rails.append(notes[time])
    out = singles.copy() if singles_mode == "anchor" else {}
    for rail_nodes in rails:
        rail_t = rail_nodes[:,2]
        a_idx = []
        b_idx = []
        last_pivot = 0
        for n, t in enumerate(rail_t):
            # rail start and end, and singles are optionally anchors
            if n in (0, len(rail_t)-1) or (singles_mode in ("anchor", "temp_anchor") and t in singles):
                a_idx.append(n)
                b_idx.append(n)
                last_pivot = n
            if (singles_mode in ("anchor", "temp_anchor") and t in singles):
                rail_nodes[n] = singles[t][0]  # snap rail to single note
            # sort even ones in the A set (moves during B), and odd ones in the B set (moves during A)
            elif (n - last_pivot) % 2 == 0:
                a_idx.append(n)
            else:
                b_idx.append(n)
        for i in range(iterations):
            # interpolate nodes that are not in the A set for even iterations or B for odd
            new = interpolate_spline(rail_nodes[a_idx if i % 2 == 0 else b_idx], rail_t)
            # weighted average, with weight <resistance> for existing nodes and 1 for interpolated ones
            rail_nodes = (rail_nodes*resistance + new)/(resistance + 1)
        # output smoothed rail
        out[rail_nodes[0,2]] = rail_nodes

    if singles_mode == "snap":
        return snap_singles_to_rail(notes = out | singles, direction=direction)
    return out


from argparse import ArgumentParser, RawDescriptionHelpFormatter
from io import BytesIO
import json
from json import JSONDecodeError
import logging
import os
from pathlib import Path
from time import strftime, time, sleep
from zipfile import ZipFile

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import PowerNorm
from matplotlib.lines import Line2D
from matplotlib.ticker import FuncFormatter
from matplotlib.widgets import Button, CheckButtons, RadioButtons, Slider
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

from . import movement, synth_format, utils, __version__
from .rails import interpolate_spline
from .synth_format import DataContainer, SynthFile

NOTE_COLORS = {"right": "red", "left": "cyan", "single": "lime", "both": "gold"}

RENDER_WINDOW = 4  # 4 seconds of elements are rendered
WALL_DESPAWN_PC = 80  # limit of visible walls, earlier ones despawn
WALL_DESPAWN_QUEST = 40
# TODO: quest delay needs more research, sometimes 47 ms delay still renders
# TODO: find a way to configure this in a way that matches ingame settings
QUEST_WALL_DELAY = 0.050  # 50 ms on highest wall density

DEFAULT_OUTPUT_FILE = Path("output.synth")  # inside working directory
DEFAULT_BACKUP_DIR = Path("smh_backup")  # inside working directory

DEFAULT_VELOCITY_WINDOW = utils.SecondFloat(0.5)  # half a second gap breaks up velocity / acceleration calculations
DEFAULT_INTERPOLATION = 1/16
DEFAULT_VEL_LIMIT = 100
DEFAULT_ACC_LIMIT = 2000
DEFAULT_HEATMAP_BIN_COUNT = 3
DEFAULT_HEATMAP_NORM_POWER = 0.6

DEFAULT_WINDOW_WIDTH = 1600
DEFAULT_WINDOW_HEIGHT = 900
DEFAULT_MENU_SIZE = 100

# TODO: make those command line arguments
AUTORELOAD_DEFAULT = True
AUTORELOAD_WAIT_SEC = 1  # Editor overwrites the file twice, we want the final result
AUTORELOAD_COOLDOWN_SEC = 5  # Don't reload very fast

AUTOBACKUP_DEFAULT_MIN = 5  # in minutes, or None

def get_parser():
    parser = ArgumentParser(
        formatter_class=RawDescriptionHelpFormatter,
        prog=f"python3 -m {__package__}.{Path(__file__).stem}",
        description='\n'.join([
            "velocity window and interpolation accept any time format, by default in beats, but also '0.5s'",
            ""
            "Also see the wiki on GitHub, which contains more detailed explainations, as well as some examples and images: https://github.com/adosikas/synth_mapping_helper/wiki",
        ]),
        epilog=f"Version: {__version__}",
    )
    parser.add_argument("input", type=Path, help="Input file")
    parser.add_argument("-o", "--output-file", type=Path, default=DEFAULT_OUTPUT_FILE, help=f"Output file. Default: {DEFAULT_OUTPUT_FILE}")
    parser.add_argument("-b", "--backup-dir", type=Path, default=DEFAULT_BACKUP_DIR, help=f"Directory for autobackups. Default: {DEFAULT_BACKUP_DIR}")

    parser.add_argument("--velocity-window", type=utils.parse_time, default=DEFAULT_VELOCITY_WINDOW, help=f"Breaks at least this long reset velocity/acceleration calculation. Default: {DEFAULT_VELOCITY_WINDOW}")
    parser.add_argument("--interpolation", type=utils.parse_time, default=DEFAULT_INTERPOLATION, help=f"Interpolation distance for plots and velocity/accelation calculation. Default: {DEFAULT_INTERPOLATION}")
    parser.add_argument("--velocity-limit", type=float, default=DEFAULT_VEL_LIMIT, help=f"Starting limit of the velocity plot. Default: {DEFAULT_VEL_LIMIT}")
    parser.add_argument("--acceleration-limit", type=float, default=DEFAULT_ACC_LIMIT, help=f"Starting limit of the acceleration plot. Default: {DEFAULT_ACC_LIMIT}")
    parser.add_argument("--heatmap-bin-count", type=float, default=DEFAULT_HEATMAP_BIN_COUNT, help=f"Number of heatmap bins per grid square. Default: {DEFAULT_HEATMAP_BIN_COUNT}")
    parser.add_argument("--heatmap-norm-power", type=float, default=DEFAULT_HEATMAP_NORM_POWER, help=f"Power of heatmap normalisation. Default: {DEFAULT_HEATMAP_NORM_POWER}")

    parser.add_argument("--window-size", type=str, default=f"{DEFAULT_WINDOW_WIDTH}x{DEFAULT_WINDOW_HEIGHT}", help=f"Size of plot window (excluding toolbars) in pixels. Default: {DEFAULT_WINDOW_WIDTH}x{DEFAULT_WINDOW_HEIGHT}")
    parser.add_argument("--menu-size", type=int, default=f"{DEFAULT_MENU_SIZE}", help=f"Height of menu in pixels. Also influences text size. Default: {DEFAULT_MENU_SIZE}")
    return parser

def abort(reason: str):
    logging.error(reason)
    exit(1)

def plot_bookmarks(bookmarks: dict[float, str], axs: "container of mpl axes"):
    for time, name in bookmarks.items():
        for ax in axs:
            ax.axvline(time, color="grey")
        axs[0].text(time, 0.99, name, ha='left', va='bottom', rotation=45, transform=axs[0].get_xaxis_transform())

def prepare_data(data: DataContainer, options) -> tuple[
    list[tuple[list["rails"], "positions", "velocity", "acceleration"]],  # note data
    list[tuple[  # wall data (pc + quest)
        tuple["pc_density", "pc_ok", "pc_despawn"],
        tuple["quest_density", "quest_ok", "quest_despawn", "quest_hidden"]
    ]],
]:
    # NOTES DATA
    velocity_window_beats = utils.second_to_beat(options.velocity_window, data.bpm)
    note_out: list[tuple[list["rails"], "positions", "velocity", "acceleration"]] = []
    for t, note_type in enumerate(synth_format.NOTE_TYPES):
        rails: list["xyz"] = []
        pos_dict: dict[float, "xyz"] = {}
        notes_dict = getattr(data, note_type)
        if not notes_dict:
            # early abort when there are no notes
            note_out.append(([], np.zeros((0,2)), np.zeros((0,2)), np.zeros((0,2))))
            continue
        
        # first, dump all notes and rails nodes into a dict, to get positions over time
        last_time = sorted(notes_dict)[0]
        for time in sorted(notes_dict):
            if time - last_time > velocity_window_beats:
                # insert None to mark end of this window
                pos_dict[(last_time+time)/2] = None
            pos = notes_dict[time]
            for p in pos:
                pos_dict[p[2]] = p[:3]
            if pos.shape[0] > 1:
                # add interpolated spline for this rail (ensuring last position matches the input, ie the interpolation doesn't make it longer)
                rails.append(interpolate_spline(pos, utils.bounded_arange(pos[0,2], pos[-1,2], options.interpolation)))
            last_time = pos[-1,2]

        pos_dict[last_time+1] = None  # cap off with None to simply the coming loop
        # convert dict to np array [time, xyz], interpolating windows at 1/64
        all_pos = []
        current_window = []
        for _, p in sorted(pos_dict.items()):
            if p is not None:
                current_window.append(p)
            else:
                if len(current_window) == 1:
                    # add single note
                    all_pos.extend(current_window)
                elif len(current_window) > 1:
                    # same as rail interpolation above, but on window instead of rail
                    all_pos.extend(interpolate_spline(np.array(current_window), utils.bounded_arange(current_window[0][2], current_window[-1][2], options.interpolation)))
                # add spacer to break up vel/acc plots
                all_pos.append((np.nan, np.nan, np.nan))
                current_window = []
        pos = np.array(all_pos)
        # velocity calculation: delta_xy / delta_t = delta_v
        # this will be plottet at each center between two notes
        delta_pos = np.diff(pos, axis=0)  # difference in x, y and time
        vel = delta_pos[:, :2] / delta_pos[:, 2:3]
        # acceleration: delta_vel / avg(preceding_delta_t, following_delta_t)²
        # this will be plotted at every note position (excluding first and last of each window)
        delta_vel = np.diff(delta_pos, axis=0)
        avg_delta_t = ((delta_pos[:-1, 2:3] + delta_pos[1:, 2:3])/2)
        acc = delta_vel[:, :2] / avg_delta_t ** 2
        note_out.append((rails, pos, vel, acc))

    wall_out: list[tuple[  # wall data (pc + quest)
        tuple["pc_density", "pc_ok", "pc_despawn"],
        tuple["quest_density", "quest_ok", "quest_despawn", "quest_hidden"]
    ]] = []
    
    # walls
    all_walls: dict[str, list[float]] = {
        w_type: sorted([t for t, wall in data.walls.items() if wall[0,3] == w])
        for w, w_type in synth_format.WALL_LOOKUP.items()
    }

    wall_markers = {
        "wall_left": ("s", "left", 7),
        "wall_right": ("s", "right", 6),
        "angle_left": ("o", "left", 5),
        "angle_right":  ("o", "right", 4),
        "center": ("d", "full", 3),
        "crouch": ("s", "top", 2),
        "triangle": ("^", "none", 1),
        "square": ("s", "none", 0),
    }
    quest_wall_delay_beats = utils.second_to_beat(QUEST_WALL_DELAY, data.bpm)

    for wall_type in synth_format.WALL_TYPES:
        pc_density: list[tuple[float, int|float]] = []
        quest_density: list[tuple[float, int|float]] = []

        pc_ok: list[float] = []
        pc_despawn: list[float] = []
        quest_hidden: list[float] = []
        quest_nothidden: list[float] = []
        quest_ok: list[float] = []
        quest_despawn: list[float] = []

        last_time_quest = None
        for i, time in enumerate(all_walls[wall_type]):
            # pc - filter out despawn
            visible_count = 1
            end = time + RENDER_WINDOW
            for other_time in all_walls[wall_type][i+1:]:
                if other_time > end:  # next wall is not visible yet
                    pc_density.append((time, visible_count))
                    if visible_count == 1:  # only a single wall visible (ie last wall of a group)
                        # break up density plot
                        pc_density.append(((time+other_time)/2, np.nan))
                    break
                visible_count += 1
            else:  # all remaining walls are visisble
                pc_density.append((time, visible_count))
            if visible_count > WALL_DESPAWN_PC:
                pc_despawn.append(time)
            else:
                pc_ok.append(time)
            # quest - first step: filter out hidden ones
            hidden_on_quest = last_time_quest is not None and time - last_time_quest < quest_wall_delay_beats
            if hidden_on_quest:
                quest_hidden.append(time)
            else:
                last_time_quest = time
                quest_nothidden.append(time)

        # quest - second step: filter out despawn
        for i, time in enumerate(quest_nothidden):
            visible_count = 1
            end = time + RENDER_WINDOW
            for other_time in quest_nothidden[i+1:]:
                if other_time > end:  # next wall is not visible yet
                    quest_density.append((time, visible_count))
                    if visible_count == 1:  # only a single wall visible (ie last wall of a group)
                        # break up density plot
                        quest_density.append(((time+other_time)/2, np.nan))
                    break
                visible_count += 1
            else:  # all remaining walls are visisble
                quest_density.append((time, visible_count))
            if visible_count > WALL_DESPAWN_QUEST:
                quest_despawn.append(time)
            else:
                quest_ok.append(time)

        wall_out.append(((pc_density, pc_ok, pc_despawn), (quest_density, quest_ok, quest_despawn, quest_hidden)))
    return note_out, wall_out

def plot_notes(options, fig, infile: SynthFile, data: DataContainer, prepared_data, **kwargs):
    axs = fig.subplots(4, sharex=True)
    axs[-1].set_xlabel("beats")
    plot_bookmarks(infile.bookmarks, axs)

    ax_x, ax_y, ax_vel, ax_acc = axs
    ax_x.set_ylabel("X-Position (sq)")
    ax_x.set_ylim((8,-8))
    ax_x.set_yticks(range(-8,8+1), (str(t) if not t%2 else "" for t in range(-8,8+1)))
    ax_x.grid(True)

    ax_y.set_ylabel("Y-Position (sq)")
    ax_y.set_ylim((-6,6))
    ax_y.set_yticks(range(-6,6+1), (str(t) if not t%2 else "" for t in range(-6,6+1)))
    ax_y.grid(True)

    ax_vel.set_ylabel("Velocity (sq/s)")
    ax_vel.set_ylim((0, options.velocity_limit))
    vel_mul = utils.second_to_beat(1, data.bpm)
    ax_acc.set_ylabel("Acceleration (sq/s²)")
    ax_acc.set_ylim((0, options.acceleration_limit))
    acc_mul = vel_mul * vel_mul

    for t, note_type in enumerate(synth_format.NOTE_TYPES):
        color = NOTE_COLORS[note_type]
        notes_dict = getattr(data, note_type)
        
        for time in sorted(notes_dict):
            pos = notes_dict[time]
            # single note / rail head
            ax_x.plot(pos[0, 2], pos[0, 0], color=color, linestyle="", marker="o")
            ax_y.plot(pos[0, 2], pos[0, 1], color=color, linestyle="", marker="o")

            if pos.shape[0] != 1:
                # rail segment markers
                ax_x.plot(pos[1:, 2], pos[1:, 0], color=color, linestyle="", marker=".")
                ax_y.plot(pos[1:, 2], pos[1:, 1], color=color, linestyle="", marker=".")

        rails, pos, vel, acc = prepared_data[0][t]
        # actual rail paths
        for r in rails:
            ax_x.plot(r[:, 2], r[:, 0], color=color)
            ax_y.plot(r[:, 2], r[:, 1], color=color)

        if pos.shape[0] > 1:
            # velocity (n-1), plotted between two nodes
            #                                                  [sq/beat] * [beat/sec] = [sq/s]
            ax_vel.plot((pos[1:, 2] + pos[:-1, 2])/2, [np.sqrt(v.dot(v)) * vel_mul for v in vel], color=color)
        if pos.shape[0] > 2:
            # acceleration (n-2) plotted at nodes (skipping first and last position)
            #                                 [sq/beat²] * ([beat/sec])² = [sq/s²]
            ax_acc.plot(pos[1:-1, 2], [np.sqrt(a.dot(a)) * acc_mul for a in acc], color=color)

def plot_walls(options, fig, infile: SynthFile, data: DataContainer, prepared_data, platform: str):
    axs = fig.subplots(2, sharex=True)
    axs[-1].set_xlabel("beats")
    plot_bookmarks(infile.bookmarks, axs)

    ax_density, ax_status = axs
    ax_density.set_ylabel("Walls visible")
    if platform == "PC":
        ax_density.set_ylim([0, 120])
        ax_density.axhline(80, color="red")
    else:
        ax_density.set_ylim([0, 60])
        ax_density.axhline(40, color="red")
    ax_status.set_ylabel("Wall Type")
    ax_status.set_yticks([])
    ax_status.set_ylim((-0.5,9))

    wall_markers = {
        # wall_type: (marker_shape, marker_fill, y-slot in type plot)
        "wall_left": ("s", "left", 7),
        "wall_right": ("s", "right", 6),
        "angle_left": ("o", "left", 5),
        "angle_right":  ("o", "right", 4),
        "center": ("d", "full", 3),
        "crouch": ("s", "top", 2),
        "triangle": ("^", "none", 1),
        "square": ("s", "none", 0),
    }
    quest_wall_delay_beats = utils.second_to_beat(QUEST_WALL_DELAY, data.bpm)

    for i, wall_type in enumerate(synth_format.WALL_TYPES):
        marker, fill, y = wall_markers[wall_type]
        if platform == "PC":
            # pc: ok first, despawn on top
            pc_density, pc_ok, pc_despawn = prepared_data[1][i][0]
            for walls, color in zip((pc_ok, pc_despawn), ("green", "orange")):
                for time in walls:
                    ax_status.plot([time], [y], marker=marker, fillstyle=fill, color=color)
            ax_density.plot([time for time, _ in pc_density], [count for _, count in pc_density], label=wall_type, marker=".")
        else:
            # quest: ok first, then despawn, hidden on top
            quest_density, quest_ok, quest_despawn, quest_hidden = prepared_data[1][i][1]
            for walls, color in zip((quest_ok, quest_despawn, quest_hidden), ("green", "orange", "red")):
                for time in walls:
                    ax_status.plot([time], [y], marker=marker, fillstyle=fill, color=color)
            ax_density.plot([time for time, _ in quest_density], [count for _, count in quest_density], label=wall_type, marker=".")


    legend_elements = [
        Line2D([0], [0], color='green', label='ok', marker="o", linestyle=""),
        Line2D([0], [0], color='orange', label='despawn', marker="o", linestyle=""),
        Line2D([0], [0], color='red', label='hidden', marker="o", linestyle=""),
    ]
    ax_density.legend(loc="upper right", ncol=len(wall_markers))
    if platform == "PC":
        ax_status.legend(handles=legend_elements[:2], loc="upper right", ncol=2)
    else:
        ax_status.legend(handles=legend_elements, loc="upper right", ncol=3)

def plot_heatmap(options, fig, infile: SynthFile, data: DataContainer, prepared_data, platform: str):
    axs = fig.subplots(2,2)
    axs = (
        axs[0,1],  # right
        axs[0,0],  # left
        axs[1,0],  # single
        axs[1,1],  # both
    )
    for t, note_type in enumerate(synth_format.NOTE_TYPES):
        _, pos, _, _ = prepared_data[0][t]
        axs[t].hist2d(
            pos[:, 0], pos[:,1],
            bins=[int(17 * options.heatmap_bin_count),int(13 * options.heatmap_bin_count)], range=[[-8,8],[-6,6]],
            cmap=("Reds", "Blues", "Greens", "Oranges")[t],
            density=False, norm=PowerNorm(options.heatmap_norm_power),
        )
        avg = np.nanmean(pos, axis=0)
        axs[t].plot(avg[0], avg[1], marker="x", color="white", markeredgecolor="black", markersize=10)
        axs[t].grid(True)
        axs[t].text((-7,7,7,-7)[t], (-5,-5,5,5)[t], len(getattr(data, note_type)), ha="center", va="center")

def show_warnings(options, fig, infile: SynthFile, data: DataContainer, prepared_data, platform: str):
    ax = fig.subplots(1)
    ax.set_axis_off()
    ax.table(loc="top", colWidths=(1/20, 19/20), cellLoc="left", rowLabels=["123 5/6"], cellText=[["note:left", "placeholder, ie extreme peak in acceleration"]])

TABS = [
    ("Notes", plot_notes),
    ("Walls", plot_walls),
    ("Heatmap", plot_heatmap),
    # ("Warnings", show_warnings)
]
PLATFORMS = ("PC", "Quest")

def main(options):
    if not options.input.is_file():
        abort("Input file is not a file, is the path correct?")

    win_sizes = options.window_size.split("x")
    if len(win_sizes) != 2:
        abort("Window size must be '<width>x<height>'")
    # scale the GUI based on one GUI block
    win_w = int(win_sizes[0]) / options.menu_size
    win_h = int(win_sizes[1]) / options.menu_size

    if win_w < 5:
        abort("Window width is too small")
    if win_h < 5:
        abort("Window height is too small")

    # load beatmap json
    data = synth_format.import_file(options.input)

    if isinstance(options.velocity_window, utils.SecondFloat):
        options.velocity_window = options.velocity_window.to_beat(data.bpm)
    if isinstance(options.interpolation, utils.SecondFloat):
        options.interpolation = options.interpolation.to_beat(data.bpm)

    fig = plt.figure(figsize=(win_w, win_h), dpi=options.menu_size, layout="constrained")
    fig.canvas.manager.set_window_title(f"SMH Companion - {data.meta.artist} - {data.meta.name}")
    menu, tab = fig.subfigures(2, 1, height_ratios=(1,win_h-1))

    active_tab = 0
    active_difficulty = 0
    active_platform = 0

    difficulties: list[tuple["difficulty_name", "prepared_data"]] = None
    def replace_prepared_data():
        nonlocal difficulties
        logging.info("Preparing data")
        difficulties = [(d, prepare_data(data.difficulties[d], options)) for d in synth_format.DIFFICULTIES if d in data.difficulties]
        if not difficulties:
            abort("No notes found")
        logging.info("Preparing complete")

    replace_prepared_data()

    def redraw():
        tab_name, tab_func = TABS[active_tab]
        diff_name, prepared_data = difficulties[active_difficulty]
        logging.info(f"Drawing {tab_name}: {diff_name}")
        tab.clear()
        tab_func(options, tab, data, data.difficulties[diff_name], prepared_data, platform=PLATFORMS[active_platform])
        fig.canvas.draw()
        fig.canvas.flush_events()
        logging.info("Drawing complete")

    btns = {}
    colors = ("darksalmon", "limegreen")
    hovercolors = ("lightsalmon", "lightgreen")
    reload_last = time()
    autobackup_last = None

    autobackup_interval_min = AUTOBACKUP_DEFAULT_MIN

    # this is changed when doing reload and autobackup
    reload_txt = menu.text(4.3/win_w, 0.3, f"Last reload: {strftime('%H:%M:%S')}", horizontalalignment="left", verticalalignment="center")
    autobackup_txt = menu.text(4.3/win_w, 0.1, "Last backup: <None>", horizontalalignment="left", verticalalignment="center")

    # this is used by btn_reload
    def btn_backup(ev) -> None:
        nonlocal autobackup_last, autobackup_interval_min
        autobackup_last = time()
        autobackup_txt.set_text(f"Last backup: {strftime('%H:%M:%S')}")
        redraw()  # force redraw to update text

        options.backup_dir.mkdir(exist_ok=True, parents=True)  # make dir if it doesn't exist
        out = options.backup_dir / f"{data.input_file.stem}_{strftime('%Y-%m-%d_%H-%M-%S')}{data.input_file.suffix}"
        out.write_bytes(data.input_file.read_bytes())  # binary identical copy, do not rexport
        logging.info(f"Backup created at {out.resolve()}")

    # this is used by autoreloader
    def btn_reload(ev) -> None:
        nonlocal data, reload_last, autobackup_last, autobackup_interval_min
        reload_last = time()
        reload_txt.set_text(f"Last reload: {strftime('%H:%M:%S')}")
        data = synth_format.import_file(options.input)
        replace_prepared_data()
        logging.info(f"Reloaded {options.input.resolve()}")
        # Update finalize button
        btns["Finalize"].label.set_text("UnFinalize" if data.bookmarks.get(0.0) == "#smh_finalized" else "Finalize")
        # Make a backup when enough time has passed
        if autobackup_interval_min is not None and (autobackup_last is None or time() > autobackup_last + autobackup_interval_min * 60):
            btn_backup(None)  # this redraws
        else:
            redraw()

    file_observer = None
    def refresh_watchdog():
        nonlocal file_observer
        if file_observer is not None:
            file_observer.stop()
        file_observer = Observer()

        class FileEventHandler(FileSystemEventHandler):
            def on_modified(self, event):
                if Path(event.src_path) == options.input:
                    if time() > reload_last + AUTORELOAD_COOLDOWN_SEC:  # don't reload twice
                        # TODO: figure out if something actually changed (only beatmap and ignore image/metadata/etc)
                        logging.info(f"Detected file modification, reloading shortly")
                        sleep(AUTORELOAD_WAIT_SEC)
                        btn_reload(None)  # simulate button click
                    refresh_watchdog()  # editor does weird stuff, so restart watchdog every time
        # Windows does not support watching single files, so we watch the whole directory (and filter for the target)
        file_observer.schedule(FileEventHandler(), options.input.parent if os.name == "nt" else options.input)
        file_observer.start()

    if AUTORELOAD_DEFAULT:
        refresh_watchdog()
    
    def btn_autoreload(ev) -> None:
        nonlocal file_observer
        if file_observer is None:
            refresh_watchdog()
            btns["AutoReload"].color = colors[True]
            btns["AutoReload"].hovercolor = hovercolors[True]
        else:
            file_observer.stop()
            file_observer = None
            btns["AutoReload"].color = colors[False]
            btns["AutoReload"].hovercolor = hovercolors[False]
        plt.draw()

    def btn_finalize(ev) -> None:
        if data.bookmarks.get(0.0) != "#smh_finalized":
            data.bookmarks[0.0] = "#smh_finalized"
            for _, diff_data in data.difficulties.items():
                diff_data.apply_for_walls(movement.offset, offset_3d=(0,-2.1,0), types=synth_format.SLIDE_TYPES)
            btns["Finalize"].label.set_text("UnFinalize")
            logging.info("Finalized map")
        elif data.bookmarks.get(0.0) == "#smh_finalized":
            del data.bookmarks[0.0]
            for _, diff_data in data.difficulties.items():
                diff_data.apply_for_walls(movement.offset, offset_3d=(0,2.1,0), types=synth_format.SLIDE_TYPES)
            btns["Finalize"].label.set_text("Finalize")
            logging.info("Reversed finalization")
        plt.draw()

    def btn_output(active: bool) -> None:
        data.save_as(options.output_file)
        logging.info(f"Saved output to {options.output_file.resolve()}")

    btn_info = [
        ("Reload", btn_reload),
        ("AutoReload", btn_autoreload),
        ("Finalize", btn_finalize),
        ("Output", btn_output),
    ]

    for i, (text, func) in enumerate(btn_info):
        btns[text] = Button(menu.add_axes((0,(len(btn_info)-(i+0.9))/len(btn_info),1/win_w,0.9/len(btn_info))), text)
        btns[text].on_clicked(func)

    btns["AutoReload"].color = colors[file_observer is not None]
    btns["AutoReload"].hovercolor = hovercolors[file_observer is not None]
    if data.bookmarks.get(0.0) == "#smh_finalized":
        btns["Finalize"].label.set_text("UnFinalize")

    tab_rb = RadioButtons(menu.add_axes((1/win_w,0,1/win_w,1)), [n for n, *_ in TABS], active_tab)
    def tab_clicked(selected: str):
        nonlocal active_tab
        for tab_id, (name, *_) in enumerate(TABS):
            if name == selected and tab_id != active_tab:
                active_tab = tab_id
                redraw()
                break
    tab_rb.on_clicked(tab_clicked)

    diff_rb = RadioButtons(menu.add_axes((2/win_w,0,1/win_w,1)), [d for d, *_ in difficulties], active_difficulty)
    def diff_clicked(selected: str):
        nonlocal active_difficulty
        for diff_id, (name, _) in enumerate(difficulties):
            if name == selected and diff_id != active_difficulty:
                active_difficulty = diff_id
                redraw()
                break
    diff_rb.on_clicked(diff_clicked)

    platform_rb = RadioButtons(menu.add_axes((3/win_w,0,1/win_w,1)), PLATFORMS, active_difficulty)
    def platform_clicked(selected: str):
        nonlocal active_platform
        for platform_id, name in enumerate(PLATFORMS):
            if name == selected and platform_id != active_platform:
                active_platform = platform_id
                redraw()
                break
    platform_rb.on_clicked(platform_clicked)

    backup_btn = Button(menu.add_axes((4/win_w,3/4,1/win_w,1/4)), "Backup now")
    backup_btn.on_clicked(btn_backup)

    autobackup_btn = Button(menu.add_axes((5/win_w,3/4,1/win_w,1/4)), "AutoBackup")
    autobackup_sl = Slider(menu.add_axes((4.05/win_w,0.6,1.9/win_w,1/8)), "", 0, 15, valinit=autobackup_interval_min or 0, valstep=1)
    autobackup_sl.valtext.set_visible(False)
    autobackup_slider_txt = menu.text(4.3/win_w, 0.5, f"Minimum age: {autobackup_interval_min} min", horizontalalignment="left", verticalalignment="center")

    def btn_autobackup(ev):
        nonlocal autobackup_interval_min
        if autobackup_interval_min is None:
            autobackup_interval_min = autobackup_sl.val
            autobackup_btn.color = colors[True]
            autobackup_btn.hovercolor = hovercolors[True]
        else:
            autobackup_interval_min = None
            autobackup_btn.color = colors[False]
            autobackup_btn.hovercolor = hovercolors[False]
    autobackup_btn.on_clicked(btn_autobackup)
    autobackup_btn.color = colors[autobackup_interval_min is not None]
    autobackup_btn.hovercolor = hovercolors[autobackup_interval_min is not None]

    def sl_autobackup(ev):
        nonlocal autobackup_interval_min
        autobackup_slider_txt.set_text(f"Minimum age: {autobackup_sl.val} min")
        if autobackup_interval_min is not None:
            autobackup_interval_min = autobackup_sl.val
    autobackup_sl.on_changed(sl_autobackup)

    redraw()
    plt.show()

def entrypoint():
    logging.basicConfig(level=logging.INFO, format='%(asctime)s [%(levelname)s] %(message)s')
    main(get_parser().parse_args())

if __name__ == "__main__":
    entrypoint()
